from typing import Any, List, Union, Dict, Type
from dataclasses import dataclass
import inspect
from .. import BaseModel
try:
    from pydantic_core import PydanticUndefined
except ImportError:
    # pydantic v1
    PydanticUndefined = None
from .ProcessorBase import ProcessorBase
from .InputFile import InputFile
from .InputFolder import InputFolder
from .OutputFile import OutputFile
from .OutputFolder import OutputFolder

@dataclass
class AppProcessorInput:
    """An input file of a processor in an app"""
    name: str
    description: str
    list: bool
    def get_spec(self):
        ret: Dict[str, Any] = {
            'name': self.name,
            'description': self.description
        }
        if self.list:
            ret['list'] = True
        return ret
    @staticmethod
    def from_spec(spec):
        return AppProcessorInput(
            name=spec['name'],
            description=spec['description'],
            list=spec.get('list', False)
        )

@dataclass
class AppProcessorInputFolder:
    """An input folder of a processor in an app"""
    name: str
    description: str
    list: bool
    def get_spec(self):
        ret: Dict[str, Any] = {
            'name': self.name,
            'description': self.description
        }
        if self.list:
            ret['list'] = True
        return ret
    @staticmethod
    def from_spec(spec):
        return AppProcessorInputFolder(
            name=spec['name'],
            description=spec['description'],
            list=spec.get('list', False)
        )

@dataclass
class AppProcessorOutput:
    """An output file of a processor in an app"""
    name: str
    description: str
    def get_spec(self):
        return {
            'name': self.name,
            'description': self.description
        }
    @staticmethod
    def from_spec(spec):
        return AppProcessorOutput(
            name=spec['name'],
            description=spec['description']
        )

@dataclass
class AppProcessorOutputFolder:
    """An output folder of a processor in an app"""
    name: str
    description: str
    def get_spec(self):
        return {
            'name': self.name,
            'description': self.description
        }
    @staticmethod
    def from_spec(spec):
        return AppProcessorOutputFolder(
            name=spec['name'],
            description=spec['description']
        )

@dataclass
class AppProcessorParameter:
    """A parameter of a processor in an app"""
    name: str
    description: str
    type: Any
    default: Any
    options: Union[List[str], List[int], List[float], None]
    secret: bool = False
    def get_spec(self):
        ret: Dict[str, Any] = {
            'name': self.name,
            'description': self.description,
            'type': _type_to_string(self.type)
        }
        if self.default != PydanticUndefined:
            ret['default'] = self.default
        if self.options is not None:
            ret['options'] = self.options
        if self.secret:
            ret['secret'] = True
        return ret
    @staticmethod
    def from_spec(spec):
        default = spec.get('default', PydanticUndefined)
        options = spec.get('options', None)
        secret = spec.get('secret', False)
        return AppProcessorParameter(
            name=spec['name'],
            description=spec['description'],
            type=_type_from_string(spec['type']),
            default=default,
            options=options,
            secret=secret
        )

@dataclass
class AppProcessorAttribute:
    """An attribute of a processor in an app"""
    name: str
    value: str
    def get_spec(self):
        return {
            'name': self.name,
            'value': self.value
        }
    @staticmethod
    def from_spec(spec):
        return AppProcessorAttribute(
            name=spec['name'],
            value=spec['value']
        )

@dataclass
class AppProcessorTag:
    """A tag of a processor in an app"""
    tag: str
    def get_spec(self):
        return {
            'tag': self.tag
        }
    @staticmethod
    def from_spec(spec):
        return AppProcessorTag(
            tag=spec['tag']
        )

class AppProcessorException(Exception):
    pass

class AppProcessor:
    """A processor in an app"""
    def __init__(self, *,
        name: str,
        description: str,
        label: str,
        inputs: List[AppProcessorInput],
        input_folders: List[AppProcessorInputFolder],
        outputs: List[AppProcessorOutput],
        output_folders: List[AppProcessorOutputFolder],
        parameters: List[AppProcessorParameter],
        attributes: List[AppProcessorAttribute],
        tags: List[AppProcessorTag],
        processor_class: Union[Type[ProcessorBase], None] = None
    ) -> None:
        if len(description) > 200:
            raise Exception('Description for processor must be less than 200 characters')

        self._name = name
        self._description = description
        self._label = label
        self._inputs = inputs
        self._input_folders = input_folders
        self._outputs = outputs
        self._output_folders = output_folders
        self._parameters = parameters
        self._attributes = attributes
        self._tags = tags
        self._processor_class = processor_class
    def get_spec(self):
        return {
            'name': self._name,
            'description': self._description,
            'label': self._label,
            'inputs': [i.get_spec() for i in self._inputs],
            'inputFolders': [i.get_spec() for i in self._input_folders],
            'outputs': [o.get_spec() for o in self._outputs],
            'outputFolders': [o.get_spec() for o in self._output_folders],
            'parameters': [p.get_spec() for p in self._parameters],
            'attributes': [a.get_spec() for a in self._attributes],
            'tags': [t.get_spec() for t in self._tags]
        }
    @staticmethod
    def from_spec(spec):
        inputs = [AppProcessorInput.from_spec(i) for i in spec['inputs']]
        input_folders = [AppProcessorInputFolder.from_spec(i) for i in spec.get('inputFolders', [])]
        outputs = [AppProcessorOutput.from_spec(o) for o in spec['outputs']]
        output_folders = [AppProcessorOutputFolder.from_spec(o) for o in spec.get('outputFolders', [])]
        parameters = [AppProcessorParameter.from_spec(p) for p in spec['parameters']]
        attributes = [AppProcessorAttribute.from_spec(a) for a in spec['attributes']]
        tags = [AppProcessorTag.from_spec(t) for t in spec['tags']]
        return AppProcessor(
            name=spec['name'],
            description=spec['description'],
            label=spec['label'],
            inputs=inputs,
            input_folders=input_folders,
            outputs=outputs,
            output_folders=output_folders,
            parameters=parameters,
            attributes=attributes,
            tags=tags
        )
    @staticmethod
    def from_processor_class(processor_class: Type[ProcessorBase]):
        name = processor_class.name
        description = processor_class.description
        label = processor_class.label
        tags = processor_class.tags
        attributes = processor_class.attributes

        _attributes: List[AppProcessorAttribute] = []
        for key, value in attributes.items():
            _attributes.append(AppProcessorAttribute(
                name=key,
                value=value
            ))
        _tags = [AppProcessorTag(tag=tag) for tag in tags]

        inputs, input_folders, outputs, output_folders, parameters = _get_context_inputs_outputs_parameters_for_processor(processor_class)

        return AppProcessor(
            name=name,
            description=description,
            label=label,
            inputs=inputs,
            input_folders=input_folders,
            outputs=outputs,
            output_folders=output_folders,
            parameters=parameters,
            attributes=_attributes,
            tags=_tags,
            processor_class=processor_class
        )

def _get_context_inputs_outputs_parameters_for_processor(processor_class: Type[ProcessorBase]):
    run_signature = inspect.signature(processor_class.run)
    run_parameters = run_signature.parameters
    if len(run_parameters) != 1:
        raise Exception('The run method should have exactly one parameter')
    context_param = list(run_parameters.values())[0]

    return _get_context_inputs_outputs_parameters_for_model(context_param.annotation)

def _get_context_inputs_outputs_parameters_for_model(context_class: Type[BaseModel]):
    context_fields = []
    try:
        # This works in pydantic v2
        model_fields = context_class.model_fields # type: ignore
    except AttributeError:
        # This is the alternative for pydantic v1
        model_fields = context_class.__fields__
    for name, field in model_fields.items():
        if name is None:
            continue
        if name == 'model_config':
            # necessary to skip for our version of pydantic BaseModel
            continue
        if hasattr(field, 'json_schema_extra'):
            json_schema_extra = field.json_schema_extra # type: ignore
            secret = json_schema_extra.get('secret', None) if json_schema_extra is not None else None # type: ignore
            assert isinstance(secret, bool) or secret is None
            options = json_schema_extra.get('options', None) if json_schema_extra is not None else None # type: ignore
        else:
            secret: Union[bool, None] = None
            options = None
        # support both pydantic v1 and v2
        annotation = field.annotation if hasattr(field, 'annotation') else None # type: ignore
        if annotation is None:
            # support pydantic v1
            annotation = _get_annotation_for_field_using_python_type_hints(context_class, name)
        if annotation is None:
            # Note: in pydantic < 1.10.0, the field.type_ picks up List[InputFile] as just InputFile, so that's why we are not using field.type_
            raise Exception(f'No type annotation for field: {name}')
        field_info = field.field_info if hasattr(field, 'field_info') else None # type: ignore
        if field_info is not None:
            description = field_info.description if hasattr(field_info, 'description') else '' # type: ignore
        else:
            description = field.description if hasattr(field, 'description') else '' # type: ignore
        context_fields.append({
            'name': name,
            'description': description,
            'annotation': annotation,
            'default': field.default if hasattr(field, 'default') else PydanticUndefined,
            'options': options,
            'secret': secret,
        })

    inputs: List[AppProcessorInput] = []
    input_folders: List[AppProcessorInputFolder] = []
    outputs: List[AppProcessorOutput] = []
    output_folders: List[AppProcessorOutputFolder] = []
    parameters: List[AppProcessorParameter] = []
    for context_field in context_fields:
        name: str = context_field['name']
        description: str = context_field['description']
        annotation: Any = context_field['annotation']
        default: Any = context_field['default']
        secret: Union[bool, None] = context_field['secret']
        options: Union[List[str], None] = context_field['options']
        if annotation == InputFile or annotation == List[InputFile]:
            is_list = annotation == List[InputFile]
            inputs.append(AppProcessorInput(
                name=name,
                description=description,
                list=is_list
            ))
            # check to make sure other fields are not set
            if options is not None:
                raise AppProcessorException(f"Input {name} has options set - only parameters can have options")
            if secret is not None:
                raise AppProcessorException(f"Input {name} has secret set - only parameters can have secret set")
            if default is not PydanticUndefined and default is not None: # None case only necessary for pydantic v1
                raise AppProcessorException(f"Input {name} has default set - only parameters can have default set")
        elif annotation == InputFolder or annotation == List[InputFolder]:
            is_list = annotation == List[InputFolder]
            input_folders.append(AppProcessorInputFolder(
                name=name,
                description=description,
                list=is_list
            ))
            # check to make sure other fields are not set
            if options is not None:
                raise AppProcessorException(f"Input folder {name} has options set - only parameters can have options")
            if secret is not None:
                raise AppProcessorException(f"Input folder {name} has secret set - only parameters can have secret set")
            if default is not PydanticUndefined and default is not None: # None case only necessary for pydantic v1
                raise AppProcessorException(f"Input folder {name} has default set - only parameters can have default set")
        elif annotation == OutputFile:
            outputs.append(AppProcessorOutput(
                name=name,
                description=description
            ))
            # check to make sure other fields are not set
            if options is not None:
                raise AppProcessorException(f"Output {name} has options set - only parameters can have options")
            if secret is not None:
                raise AppProcessorException(f"Output {name} has secret set - only parameters can have secret set")
            if default is not PydanticUndefined and default is not None: # None case only necessary for pydantic v1
                raise AppProcessorException(f"Output {name} has default set - only parameters can have default set")
        elif annotation == OutputFolder:
            output_folders.append(AppProcessorOutputFolder(
                name=name,
                description=description
            ))
            # check to make sure other fields are not set
            if options is not None:
                raise AppProcessorException(f"Output folder {name} has options set - only parameters can have options")
            if secret is not None:
                raise AppProcessorException(f"Output folder {name} has secret set - only parameters can have secret set")
            if default is not PydanticUndefined and default is not None:
                raise AppProcessorException(f"Output folder {name} has default set - only parameters can have default set")
        elif _is_valid_parameter_type(annotation):
            parameters.append(AppProcessorParameter(
                name=name,
                description=description,
                type=annotation,
                default=default,
                options=options,
                secret=secret if secret is not None else False
            ))
        elif _is_pydantic_model_class(annotation):
            inputs0, input_folders0, outputs0, output_folders0, parameters0 = _get_context_inputs_outputs_parameters_for_model(annotation)
            for input0 in inputs0:
                input0.name = f'{name}.{input0.name}'
                inputs.append(input0)
            for input_folder0 in input_folders0:
                input_folder0.name = f'{name}.{input_folder0.name}'
                input_folders.append(input_folder0)
            for output0 in outputs0:
                output0.name = f'{name}.{output0.name}'
                outputs.append(output0)
            for output_folder0 in output_folders0:
                output_folder0.name = f'{name}.{output_folder0.name}'
                output_folders.append(output_folder0)
            for parameter0 in parameters0:
                parameter0.name = f'{name}.{parameter0.name}'
                parameters.append(parameter0)
        else:
            raise AppProcessorException(f"Unsupported type for {name}: {annotation}")
    return inputs, input_folders, outputs, output_folders, parameters

def _is_pydantic_model_class(type: Any):
    return inspect.isclass(type) and issubclass(type, BaseModel)

_type_map = {
    'str': str,
    'int': int,
    'float': float,
    'bool': bool,
    'List[str]': List[str],
    'List[int]': List[int],
    'List[float]': List[float],
    'List[bool]': List[bool],
    'Optional[int]': Union[int, None],
    'Optional[float]': Union[float, None],
    'Optional[List[str]]': Union[List[str], None],
    'Optional[List[int]]': Union[List[int], None],
    'Optional[List[float]]': Union[List[float], None],
    'Optional[List[bool]]': Union[List[bool], None]
}

def _type_to_string(type: Any):
    for key, value in _type_map.items():
        if value == type:
            return key
    raise ValueError(f'Unexpected type: {type}')

def _type_from_string(type: str):
    try:
        return _type_map[type]
    except KeyError as exc:
        raise ValueError(f'Unexpected type: {type}') from exc

def _is_valid_parameter_type(type: Any):
    return type in _type_map.values()

def _get_annotation_for_field_using_python_type_hints(model_class: Type[BaseModel], field_name: str):
    # This is a workaround for pydantic v1 where field.annotation is not set and field.type_ doesn't cut it for List[InputFile] when pydantic version is < 1.10.0
    from typing import get_type_hints
    type_hints = get_type_hints(model_class)
    return type_hints.get(field_name, None)



import os
import pytest
import time
import tempfile
import shutil
import json
import yaml


@pytest.mark.asyncio
@pytest.mark.api
async def test_integration(tmp_path):
    # important to put the imports inside so we don't get an import error when running the non-api tests
    from dendro.common.dendro_types import DendroProjectUser, DendroComputeResourceApp, CreateJobRequestInputParameter, CreateJobRequestInputFile, CreateJobRequestOutputFile
    from dendro.api_helpers.routers.gui._authenticate_gui_request import _create_mock_github_access_token
    from dendro.api_helpers.routers.gui.create_job_route import CreateJobRequest, CreateJobResponse
    from dendro.compute_resource.register_compute_resource import register_compute_resource
    from dendro.compute_resource.start_compute_resource import start_compute_resource
    from dendro.common._api_request import _use_api_test_client
    from dendro.mock import set_use_mock
    from dendro.api_helpers.clients._get_mongo_client import _clear_mock_mongo_databases
    from dendro.common._api_request import _gui_post_api_request, _client_get_api_request
    from dendro.common.dendro_types import DendroJobRequiredResources
    from dendro.api_helpers.routers.gui.job_routes import ApproveJobResponse

    os.environ['SLURM_PARTITION'] = 'default'
    os.environ['SLURM_GPU_PARTITION'] = 'default'

    tmpdir = str(tmp_path)

    from fastapi.testclient import TestClient
    app = _get_fastapi_app()
    test_client = TestClient(app)
    _use_api_test_client(test_client)
    set_use_mock(True)
    github_access_token = _create_mock_github_access_token()
    github_access_token_for_other_user = _create_mock_github_access_token()
    github_access_token_for_admin_user = _create_mock_github_access_token()
    admin_user_id = 'github|' + github_access_token_for_admin_user[len('mock:'):]

    this_dir = os.path.dirname(os.path.realpath(__file__))

    old_env = os.environ.copy()
    os.environ['ADMIN_USER_IDS'] = f'["{admin_user_id}"]'

    try:
        # Copy mock app source code
        shutil.copytree(src=this_dir + '/mock_app', dst=tmpdir + '/mock_app')
        shutil.copytree(src=this_dir + '/mock_app_2', dst=tmpdir + '/mock_app_2')

        # Create spec.json for mock_app and mock_app_2
        compute_resource_spec_app = _create_spec_json_for_mock_app(tmpdir + '/mock_app')
        compute_resource_spec_app_2 = _create_spec_json_for_mock_app(tmpdir + '/mock_app_2')

        # Create compute resource in a directory
        compute_resource_dir = tmpdir + '/compute_resource'
        os.mkdir(compute_resource_dir)
        compute_resource_id, compute_resource_private_key = register_compute_resource(dir=compute_resource_dir)
        config_fname = compute_resource_dir + '/.dendro-compute-resource-node.yaml'
        assert os.path.exists(config_fname)
        with open(config_fname, 'r') as f:
            config = yaml.safe_load(f)
        config['AVAILABLE_JOB_RUN_METHODS'] = 'local,slurm'
        with open(config_fname, 'w') as f:
            yaml.dump(config, f)

        # gui: Register compute resource
        _register_compute_resource(compute_resource_id=compute_resource_id, compute_resource_private_key=compute_resource_private_key, github_access_token=github_access_token, name='test-cr')

        # gui: Register the same compute resource again (should be okay)
        _register_compute_resource(compute_resource_id=compute_resource_id, compute_resource_private_key=compute_resource_private_key, github_access_token=github_access_token, name='test-cr-again')

        # gui: Fail register compute resource
        with pytest.raises(Exception):
            _register_compute_resource(compute_resource_id=compute_resource_id, compute_resource_private_key=compute_resource_private_key, github_access_token=github_access_token, name='test-cr', bad_resource_code=True)
        with pytest.raises(Exception):
            _register_compute_resource(compute_resource_id=compute_resource_id, compute_resource_private_key=compute_resource_private_key, github_access_token=github_access_token, name='test-cr', bad_signature=True)
        with pytest.raises(Exception):
            _register_compute_resource(compute_resource_id=compute_resource_id, compute_resource_private_key=compute_resource_private_key, github_access_token=github_access_token, name='test-cr', bad_resource_code_timestamp=True)

        # gui: set compute resource apps
        apps = [
            DendroComputeResourceApp(
                name='mock_app',
                specUri=f'file://{tmpdir}/mock_app/spec.json'
            ),
            DendroComputeResourceApp(
                name='mock_app_2',
                specUri=f'file://{tmpdir}/mock_app_2/spec.json'
            )
        ]
        _set_compute_resource_apps(
            compute_resource_id=compute_resource_id,
            apps=apps,
            github_access_token=github_access_token
        )

        # gui: Fail at setting apps by other user
        with pytest.raises(Exception):
            _set_compute_resource_apps(
                compute_resource_id=compute_resource_id,
                apps=apps,
                github_access_token=github_access_token_for_other_user
            )

        # gui: Get compute resources for user
        _check_num_compute_resources_for_user(github_access_token=github_access_token, num_compute_resources=1)

        # gui: Try to get compute resource and one that does not exist
        _get_compute_resource(compute_resource_id=compute_resource_id, github_access_token=github_access_token)
        with pytest.raises(Exception):
            _get_compute_resource(compute_resource_id='compute-resource-does-not-exist', github_access_token=github_access_token)

        # gui: Get pubsub subscription
        _get_pubsub_subscription(compute_resource_id=compute_resource_id, github_access_token=github_access_token)

        # gui: Create projects
        project1_id = _create_project('project1', github_access_token=github_access_token)
        project2_id = _create_project('project2', github_access_token=github_access_token)

        # gui: Set project name
        _set_project_name(project_id=project1_id, name='project1_renamed', github_access_token=github_access_token)

        # gui: Set project description
        _set_project_description(project_id=project1_id, description='project1_description', github_access_token=github_access_token)

        # gui: Set project tags
        _set_project_tags(project_id=project1_id, tags=['tag1', 'tag2'], github_access_token=github_access_token)

        # gui: Fail because not authenticated
        with pytest.raises(Exception):
            _set_project_name(project_id=project1_id, name='project1_renamed_2', github_access_token='')
        with pytest.raises(Exception):
            _set_project_name(project_id=project1_id, name='project1_renamed_2', github_access_token=github_access_token + '-bad') # keep the mock: at the beginning

        # gui: Get project
        project = _get_project(project_id=project1_id, github_access_token=github_access_token)
        assert project.name == 'project1_renamed'
        assert project.description == 'project1_description'
        assert project.ownerId == 'github|' + github_access_token[len('mock:'):]
        assert project.users == []
        assert project.publiclyReadable is True
        assert project.tags == ['tag1', 'tag2']
        assert project.timestampCreated > 0
        assert project.timestampModified > 0
        assert project.computeResourceId is None

        # gui: Try to get project that does not exist
        with pytest.raises(Exception):
            _get_project(project_id='project-does-not-exist', github_access_token=github_access_token)

        # client: Get project
        _client_get_project(project_id=project1_id)

        # client: Try to get project that does not exist
        with pytest.raises(Exception):
            resp = _client_get_api_request(url_path='/api/client/projects/does_not_exist')

        # gui: Set the project to be not publicly readable
        _set_project_publicly_readable(project_id=project1_id, publicly_readable=False, github_access_token=github_access_token)

        # gui: Set projects compute resource id
        for project_id in [project1_id, project2_id]:
            _set_project_compute_resource_id(project_id=project_id, compute_resource_id=compute_resource_id, github_access_token=github_access_token)

        # gui: Set project users
        users = [
            DendroProjectUser(userId='github|user_viewer', role='viewer'),
            DendroProjectUser(userId='github|user_editor', role='editor'),
            DendroProjectUser(userId='github|user_admin', role='admin')
        ]
        _set_project_users(project_id=project1_id, users=users, github_access_token=github_access_token)

        # gui: Get project
        project = _get_project(project_id=project1_id, github_access_token=github_access_token)
        assert project.publiclyReadable is False
        assert project.computeResourceId == compute_resource_id
        assert project.users == users

        # gui: Get all projects
        projects = _get_all_projects_for_user(github_access_token=github_access_token)
        assert len(projects) == 2
        assert project1_id in [p.projectId for p in projects]
        assert project2_id in [p.projectId for p in projects]

        # gui: Admin get all projects
        projects = _admin_get_all_projects(github_access_token=github_access_token_for_admin_user)
        assert len(projects) == 2

        # gui: Get projects with tag
        projects = _get_projects_with_tag(tag='tag1', github_access_token=github_access_token)
        assert len(projects) == 1
        assert project1_id in [p.projectId for p in projects]

        # gui: Delete project
        _delete_project(project_id=project1_id, github_access_token=github_access_token)

        # gui: Get all projects
        projects = _get_all_projects_for_user(github_access_token=github_access_token)
        assert len(projects) == 1
        assert project2_id in [p.projectId for p in projects]

        # gui: Create a dummy input file (required for test job)
        _create_project_file(project_id=project2_id, file_name='mock-input', content='url:https://fake-url', github_access_token=github_access_token)

        # gui: Get file
        file = _get_project_file(project_id=project2_id, file_name='mock-input', github_access_token=github_access_token)
        assert file.fileName == 'mock-input'

        # gui: Try to get file that does not exist
        with pytest.raises(Exception):
            _get_project_file(project_id=project2_id, file_name='does_not_exist', github_access_token=github_access_token)

        # client: Get project files
        files = _client_get_project_files(project_id=project2_id)
        assert len(files) == 1

        # gui: Create and delete a file
        _create_project_file(project_id=project2_id, file_name='file-to-delete.txt', content='url:https://fake-url', github_access_token=github_access_token)
        _delete_project_file(project_id=project2_id, file_name='file-to-delete.txt', github_access_token=github_access_token)

        # gui: Try to delete a file that does not exist
        with pytest.raises(Exception):
            _delete_project_file(project_id=project2_id, file_name='does_not_exist', github_access_token=github_access_token)

        # gui: Get files
        files = _get_project_files(project_id=project2_id, github_access_token=github_access_token)
        assert len(files) == 1

        # gui: Create jobs for app 1
        job_id_1 = ''
        job_id_1_with_error = ''
        processor_name = 'mock-processor1'
        processor_spec = compute_resource_spec_app.processors[0]
        for intentional_error in [False, True]:
            req = CreateJobRequest(
                projectId=project2_id,
                processorName=processor_name,
                inputFiles=[
                    CreateJobRequestInputFile(name='input_file', fileName='mock-input'),
                    CreateJobRequestInputFile(name='input_list[0]', fileName='mock-input'),
                ],
                outputFiles=[
                    CreateJobRequestOutputFile(name='output_file', fileName='mock-output' if not intentional_error else 'mock-output-err')
                ],
                inputParameters=[
                    CreateJobRequestInputParameter(name='text1', value='this is text1'),
                    CreateJobRequestInputParameter(name='text2', value='this is text2'),
                    # text3 has a default
                    CreateJobRequestInputParameter(name='val1', value=12),
                    CreateJobRequestInputParameter(name='group.num', value=3),
                    CreateJobRequestInputParameter(name='group.secret_param', value='456'),
                    CreateJobRequestInputParameter(name='intentional_error', value=intentional_error)
                ],
                processorSpec=processor_spec,
                batchId=None,
                dandiApiKey=None,
                requiredResources=DendroJobRequiredResources(numCpus=1, numGpus=0, memoryGb=8, timeSec=3600),
                runMethod='local'
            )
            resp = _gui_post_api_request(url_path='/api/gui/jobs', data=_model_dump(req), github_access_token=github_access_token)
            resp = CreateJobResponse(**resp)
            assert resp.success
            assert resp.jobId
            if not intentional_error:
                job_id_1 = resp.jobId
            else:
                job_id_1_with_error = resp.jobId

        if job_id_1:
            resp = _gui_post_api_request(url_path=f'/api/gui/jobs/{job_id_1}/approve', data={}, github_access_token=github_access_token)
            resp = ApproveJobResponse(**resp)
            assert resp.success

        if job_id_1_with_error:
            resp = _gui_post_api_request(url_path=f'/api/gui/jobs/{job_id_1_with_error}/approve', data={}, github_access_token=github_access_token)
            resp = ApproveJobResponse(**resp)
            assert resp.success

        # gui: Create job for app 2 (which uses slurm)
        processor_name_2 = 'mock-processor2'
        processor_spec_2 = compute_resource_spec_app_2.processors[0]
        req = CreateJobRequest(
            projectId=project2_id,
            processorName=processor_name_2,
            inputFiles=[],
            outputFiles=[],
            inputParameters=[
                CreateJobRequestInputParameter(
                    name='text1',
                    value='this is text1'
                )
            ],
            processorSpec=processor_spec_2,
            batchId=None,
            dandiApiKey=None,
            requiredResources=DendroJobRequiredResources(numCpus=1, numGpus=0, memoryGb=8, timeSec=3600),
            runMethod='slurm'
        )
        resp = _gui_post_api_request(url_path='/api/gui/jobs', data=_model_dump(req), github_access_token=github_access_token)
        resp = CreateJobResponse(**resp)
        assert resp.success
        job_id_2 = resp.jobId
        assert job_id_2

        if job_id_2:
            resp = _gui_post_api_request(url_path=f'/api/gui/jobs/{job_id_2}/approve', data={}, github_access_token=github_access_token)
            resp = ApproveJobResponse(**resp)
            assert resp.success

        # gui: Test not providing a required parameter
        processor_name_2 = 'mock-processor2'
        processor_spec_2 = compute_resource_spec_app_2.processors[0]
        req = CreateJobRequest(
            projectId=project2_id,
            processorName=processor_name_2,
            inputFiles=[],
            outputFiles=[],
            inputParameters=[
            ],
            processorSpec=processor_spec_2,
            batchId=None,
            dandiApiKey=None,
            requiredResources=DendroJobRequiredResources(numCpus=1, numGpus=0, memoryGb=8, timeSec=3600),
            runMethod='local'
        )
        with pytest.raises(Exception):
            _gui_post_api_request(url_path='/api/gui/jobs', data=_model_dump(req), github_access_token=github_access_token)

        # gui: Get job
        job = _get_job(job_id=job_id_1, github_access_token=github_access_token)
        assert job.projectId == project2_id
        assert job.processorName == processor_name
        assert job.processorSpec == processor_spec
        assert job.batchId is None
        assert job.dandiApiKey is None
        assert job.status == 'pending'
        assert job.timestampCreated > 0
        assert job.timestampStarted is None
        assert job.timestampFinished is None
        assert job.timestampQueued is None
        assert job.timestampStarting is None
        assert job.computeResourceId == compute_resource_id
        assert not job.jobPrivateKey # should not be exposed to GUI

        # gui: Try to get job that does not exist
        with pytest.raises(Exception):
            _get_job(job_id='job-does-not-exist', github_access_token=github_access_token)

        # gui: Get jobs
        jobs = _get_jobs(project_id=project2_id, github_access_token=github_access_token)
        assert len(jobs) == 3

        # gui: Get compute resource jobs
        jobs = _get_compute_resource_jobs(compute_resource_id=compute_resource_id, github_access_token=github_access_token)
        assert len(jobs) == 3

        # gui: Fail getting compute resource jobs by unauthorized user
        with pytest.raises(Exception):
            _get_compute_resource_jobs(compute_resource_id=compute_resource_id, github_access_token=github_access_token_for_other_user)

        # client: Get project jobs
        jobs = _client_get_project_jobs(project_id=project2_id)
        assert len(jobs) == 3

        # compute resource: get unfinished jobs
        jobs = _compute_resource_get_unfinished_jobs(compute_resource_id=compute_resource_id, compute_resource_private_key=compute_resource_private_key)
        assert len(jobs) == 3
        job = jobs[0]
        assert job.jobPrivateKey

        # compute resource: fail getting unfinished jobs
        with pytest.raises(Exception):
            _compute_resource_get_unfinished_jobs(compute_resource_id=compute_resource_id, compute_resource_private_key=compute_resource_private_key, wrong_payload=True)
        with pytest.raises(Exception):
            _compute_resource_get_unfinished_jobs(compute_resource_id=compute_resource_id, compute_resource_private_key=compute_resource_private_key, wrong_signature=True)

        # processor: Get job output upload url
        url = _get_job_output_upload_url(job_id=job.jobId, job_private_key=job.jobPrivateKey, output_name='output_file')
        assert url

        # Not enforcing this for now
        # # processor: Fail getting job output upload url for unknown output
        # with pytest.raises(Exception):
        #     _get_job_output_upload_url(job_id=job.jobId, job_private_key=job.jobPrivateKey, output_name='unknown-output')

        # processor: Fail getting job output upload url do to incorrect private key
        with pytest.raises(Exception):
            _get_job_output_upload_url(job_id=job.jobId, job_private_key=job.jobPrivateKey + 'x', output_name='output_file')

        # Run the compute resource briefly (will handle the jobs)
        start_compute_resource(dir=compute_resource_dir, timeout=3, cleanup_old_jobs=False)

        # gui: Check that first job succeeded
        job = _get_job(job_id=job_id_1, github_access_token=github_access_token)
        assert job.status == 'completed'

        # gui: Get output file for first job
        file = _get_project_file(project_id=project2_id, file_name='mock-output', github_access_token=github_access_token)
        assert file.fileName == 'mock-output'

        # gui: Check that second job failed
        job = _get_job(job_id=job_id_1_with_error, github_access_token=github_access_token)
        assert job.status == 'failed'

        # gui: Check that job from slurm app succeeded
        job = _get_job(job_id=job_id_2, github_access_token=github_access_token)
        assert job.status == 'completed'

        # Check whether the appropriate compute resource spec was uploaded to the api
        compute_resource = _get_compute_resource(compute_resource_id=compute_resource_id, github_access_token=github_access_token)
        assert compute_resource.spec
        assert len(compute_resource.spec.apps) == 2

        # gui: Try delete job without proper github access token
        with pytest.raises(Exception):
            _delete_job(job_id=job_id_1, github_access_token='bad_access_token')

        # gui: Delete job (should trigger deletion of output file)
        _delete_job(job_id=job_id_1, github_access_token=github_access_token)

        # gui: Get jobs
        jobs = _get_jobs(project_id=project2_id, github_access_token=github_access_token)
        assert len(jobs) == 2

        # gui: Get files
        files = _get_project_files(project_id=project2_id, github_access_token=github_access_token)
        assert len(files) == 2
        assert files[0].fileName == 'mock-input'

        # gui: Fail at deleting compute resource by unauthorized user
        with pytest.raises(Exception):
            _delete_compute_resource(compute_resource_id=compute_resource_id, github_access_token=github_access_token_for_other_user)

        # gui: Delete compute resource
        _delete_compute_resource(compute_resource_id=compute_resource_id, github_access_token=github_access_token)

        # gui Get compute resources
        compute_resources = _get_compute_resources(github_access_token=github_access_token)
        assert len(compute_resources) == 0
    finally:
        _use_api_test_client(None)
        set_use_mock(False)
        _clear_mock_mongo_databases()
        os.environ = old_env

def _get_fastapi_app():
    from fastapi import FastAPI

    # this code is duplicated with api/index.py, I know
    from dendro.api_helpers.routers.processor.router import router as processor_router
    from dendro.api_helpers.routers.compute_resource.router import router as compute_resource_router
    from dendro.api_helpers.routers.client.router import router as client_router
    from dendro.api_helpers.routers.gui.router import router as gui_router

    app = FastAPI()

    # requests from a processing job
    app.include_router(processor_router, prefix="/api/processor", tags=["Processor"])

    # requests from a compute resource
    app.include_router(compute_resource_router, prefix="/api/compute_resource", tags=["Compute Resource"])

    # requests from a client (usually Python)
    app.include_router(client_router, prefix="/api/client", tags=["Client"])

    # requests from the GUI
    app.include_router(gui_router, prefix="/api/gui", tags=["GUI"])

    return app

class TemporaryDirectory:
    """A context manager for temporary directories"""
    def __init__(self):
        self._dir = None
    def __enter__(self):
        self._dir = tempfile.mkdtemp()
        return self._dir
    def __exit__(self, exc_type, exc_value, traceback):
        if self._dir:
            shutil.rmtree(self._dir)

def _create_spec_json_for_mock_app(app_dir: str):
    from dendro.sdk._make_spec_file import make_app_spec_file_function
    from dendro.common.dendro_types import ComputeResourceSpecApp

    spec_fname = app_dir + '/spec.json'
    make_app_spec_file_function(app_dir=app_dir, spec_output_file=spec_fname)
    assert os.path.exists(spec_fname)
    with open(spec_fname, 'r') as f:
        spec = json.load(f)
    compute_resource_spec = ComputeResourceSpecApp(**spec)
    return compute_resource_spec

def _register_compute_resource(compute_resource_id: str, compute_resource_private_key: str, github_access_token: str, name: str, bad_resource_code_timestamp: bool = False, bad_resource_code: bool = False, bad_signature: bool = False):
    from dendro.common._crypto_keys import sign_message
    from dendro.api_helpers.routers.gui.compute_resource_routes import RegisterComputeResourceRequest, RegisterComputeResourceResponse
    from dendro.common._api_request import _gui_post_api_request

    timestamp = int(time.time())
    if bad_resource_code_timestamp:
        timestamp = timestamp - 10000
    resource_code_payload = {'timestamp': timestamp}
    resource_code_signature = sign_message(resource_code_payload, compute_resource_id, compute_resource_private_key)
    if bad_signature:
        resource_code_signature = 'bad-signature'
    resource_code = f'{resource_code_payload["timestamp"]}-{resource_code_signature}'
    if bad_resource_code:
        resource_code = 'bad-resource-code'
    req = RegisterComputeResourceRequest(
        name=name,
        computeResourceId=compute_resource_id,
        resourceCode=resource_code
    )
    resp = _gui_post_api_request(url_path='/api/gui/compute_resources/register', data=_model_dump(req), github_access_token=github_access_token)
    resp = RegisterComputeResourceResponse(**resp)
    assert resp.success

def _set_compute_resource_apps(compute_resource_id: str, apps: list, github_access_token: str):
    from dendro.api_helpers.routers.gui.compute_resource_routes import SetComputeResourceAppsRequest, SetComputeResourceAppsResponse
    from dendro.common._api_request import _gui_put_api_request

    req = SetComputeResourceAppsRequest(
        apps=apps
    )
    resp = _gui_put_api_request(url_path=f'/api/gui/compute_resources/{compute_resource_id}/apps', data=_model_dump(req), github_access_token=github_access_token)
    resp = SetComputeResourceAppsResponse(**resp)
    assert resp.success

def _check_num_compute_resources_for_user(github_access_token: str, num_compute_resources: int):
    from dendro.api_helpers.routers.gui.compute_resource_routes import GetComputeResourcesResponse
    from dendro.common._api_request import _gui_get_api_request

    resp = _gui_get_api_request(url_path='/api/gui/compute_resources', github_access_token=github_access_token)
    resp = GetComputeResourcesResponse(**resp)
    assert resp.success
    assert len(resp.computeResources) == num_compute_resources

def _get_compute_resource(compute_resource_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.compute_resource_routes import GetComputeResourceResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path=f'/api/gui/compute_resources/{compute_resource_id}', github_access_token=github_access_token)
    resp = GetComputeResourceResponse(**resp)
    assert resp.success
    assert resp.computeResource.computeResourceId == compute_resource_id
    return resp.computeResource

def _get_pubsub_subscription(compute_resource_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.compute_resource_routes import GetPubsubSubscriptionResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path=f'/api/gui/compute_resources/{compute_resource_id}/pubsub_subscription', github_access_token=github_access_token)
    resp = GetPubsubSubscriptionResponse(**resp)
    assert resp.success

def _create_project(name: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import CreateProjectRequest, CreateProjectResponse
    from dendro.common._api_request import _gui_post_api_request
    req = CreateProjectRequest(
        name='project1'
    )
    resp = _gui_post_api_request(url_path='/api/gui/projects', data=_model_dump(req), github_access_token=github_access_token)
    resp = CreateProjectResponse(**resp)
    assert resp.success
    project1_id = resp.projectId
    assert project1_id
    return project1_id

def _set_project_name(project_id: str, name: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import SetProjectNameRequest, SetProjectNameResponse
    from dendro.common._api_request import _gui_put_api_request
    req = SetProjectNameRequest(
        name='project1_renamed'
    )
    resp = _gui_put_api_request(url_path=f'/api/gui/projects/{project_id}/name', data=_model_dump(req), github_access_token=github_access_token)
    resp = SetProjectNameResponse(**resp)
    assert resp.success

def _set_project_description(project_id: str, description: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import SetProjectDescriptionRequest, SetProjectDescriptionResponse
    from dendro.common._api_request import _gui_put_api_request
    req = SetProjectDescriptionRequest(
        description='project1_description'
    )
    resp = _gui_put_api_request(url_path=f'/api/gui/projects/{project_id}/description', data=_model_dump(req), github_access_token=github_access_token)
    resp = SetProjectDescriptionResponse(**resp)
    assert resp.success

def _set_project_tags(project_id: str, tags: list, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import SetProjectTagsRequest, SetProjectTagsResponse
    from dendro.common._api_request import _gui_put_api_request
    req = SetProjectTagsRequest(
        tags=['tag1', 'tag2']
    )
    resp = _gui_put_api_request(url_path=f'/api/gui/projects/{project_id}/tags', data=_model_dump(req), github_access_token=github_access_token)
    resp = SetProjectTagsResponse(**resp)
    assert resp.success

def _get_project(project_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import GetProjectResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path=f'/api/gui/projects/{project_id}', github_access_token=github_access_token)
    resp = GetProjectResponse(**resp)
    project = resp.project
    assert project.projectId == project_id
    return project

def _client_get_project(project_id: str):
    from dendro.api_helpers.routers.client.router import GetProjectResponse as ClientGetProjectResponse
    from dendro.common._api_request import _client_get_api_request
    resp = _client_get_api_request(url_path=f'/api/client/projects/{project_id}')
    resp = ClientGetProjectResponse(**resp)
    assert resp.success
    assert resp.project.projectId == project_id

def _set_project_publicly_readable(project_id: str, publicly_readable: bool, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import SetProjectPubliclyReadableRequest, SetProjectPubliclyReadableResponse
    from dendro.common._api_request import _gui_put_api_request
    req = SetProjectPubliclyReadableRequest(
        publiclyReadable=publicly_readable
    )
    resp = _gui_put_api_request(url_path=f'/api/gui/projects/{project_id}/publicly_readable', data=_model_dump(req), github_access_token=github_access_token)
    resp = SetProjectPubliclyReadableResponse(**resp)
    assert resp.success

def _set_project_compute_resource_id(project_id: str, compute_resource_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import SetProjectComputeResourceIdRequest, SetProjectComputeResourceIdResponse
    from dendro.common._api_request import _gui_put_api_request
    req = SetProjectComputeResourceIdRequest(
        computeResourceId=compute_resource_id
    )
    resp = _gui_put_api_request(url_path=f'/api/gui/projects/{project_id}/compute_resource_id', data=_model_dump(req), github_access_token=github_access_token)
    resp = SetProjectComputeResourceIdResponse(**resp)
    assert resp.success

def _set_project_users(project_id: str, users: list, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import SetProjectUsersRequest, SetProjectUsersResponse
    from dendro.common._api_request import _gui_put_api_request
    req = SetProjectUsersRequest(
        users=users
    )
    resp = _gui_put_api_request(url_path=f'/api/gui/projects/{project_id}/users', data=_model_dump(req), github_access_token=github_access_token)
    resp = SetProjectUsersResponse(**resp)
    assert resp.success

def _get_all_projects_for_user(github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import GetProjectsResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path='/api/gui/projects', github_access_token=github_access_token)
    resp = GetProjectsResponse(**resp)
    projects = resp.projects
    return projects

def _admin_get_all_projects(github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import AdminGetAllProjectsResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path='/api/gui/projects/admin/get_all_projects', github_access_token=github_access_token)
    resp = AdminGetAllProjectsResponse(**resp)
    projects = resp.projects
    return projects

def _get_projects_with_tag(tag: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import GetProjectsResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path='/api/gui/projects?tag=tag1', github_access_token=github_access_token)
    resp = GetProjectsResponse(**resp)
    projects = resp.projects
    return projects

def _delete_project(project_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import DeleteProjectResponse
    from dendro.common._api_request import _gui_delete_api_request
    resp = _gui_delete_api_request(url_path=f'/api/gui/projects/{project_id}', github_access_token=github_access_token)
    resp = DeleteProjectResponse(**resp)
    assert resp.success

def _create_project_file(project_id: str, file_name: str, content: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.file_routes import SetFileRequest, SetFileResponse
    from dendro.common._api_request import _gui_put_api_request
    req = SetFileRequest(
        content=content,
        size=1
    )
    resp = _gui_put_api_request(url_path=f'/api/gui/projects/{project_id}/files/{file_name}', data=_model_dump(req), github_access_token=github_access_token)
    resp = SetFileResponse(**resp)
    assert resp.success

def _get_project_file(project_id: str, file_name: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.file_routes import GetFileResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path=f'/api/gui/projects/{project_id}/files/{file_name}', github_access_token=github_access_token)
    resp = GetFileResponse(**resp)
    assert resp.success
    return resp.file

def _client_get_project_files(project_id: str):
    from dendro.api_helpers.routers.client.router import GetProjectFilesResponse as ClientGetProjectFilesResponse
    from dendro.common._api_request import _client_get_api_request
    resp = _client_get_api_request(url_path=f'/api/client/projects/{project_id}/files')
    resp = ClientGetProjectFilesResponse(**resp)
    assert resp.success
    return resp.files

def _delete_project_file(project_id: str, file_name: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.file_routes import DeleteFileResponse
    from dendro.common._api_request import _gui_delete_api_request
    resp = _gui_delete_api_request(url_path=f'/api/gui/projects/{project_id}/files/{file_name}', github_access_token=github_access_token)
    resp = DeleteFileResponse(**resp)
    assert resp.success

def _get_project_files(project_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.file_routes import GetFilesResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path=f'/api/gui/projects/{project_id}/files', github_access_token=github_access_token)
    resp = GetFilesResponse(**resp)
    assert resp.success
    return resp.files

def _get_job(job_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.job_routes import GetJobResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path=f'/api/gui/jobs/{job_id}', github_access_token=github_access_token)
    resp = GetJobResponse(**resp)
    job = resp.job
    assert job.jobId == job_id
    return job

def _get_jobs(project_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.project_routes import GetJobsResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path=f'/api/gui/projects/{project_id}/jobs', github_access_token=github_access_token)
    resp = GetJobsResponse(**resp)
    jobs = resp.jobs
    return jobs

def _get_compute_resource_jobs(compute_resource_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.compute_resource_routes import GetJobsForComputeResourceResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path=f'/api/gui/compute_resources/{compute_resource_id}/jobs', github_access_token=github_access_token)
    resp = GetJobsForComputeResourceResponse(**resp)
    assert resp.success
    return resp.jobs

def _client_get_project_jobs(project_id: str):
    from dendro.api_helpers.routers.client.router import GetProjectJobsResponse as ClientGetProjectJobsResponse
    from dendro.common._api_request import _client_get_api_request
    resp = _client_get_api_request(url_path=f'/api/client/projects/{project_id}/jobs')
    resp = ClientGetProjectJobsResponse(**resp)
    assert resp.success
    return resp.jobs

def _delete_job(job_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.job_routes import DeleteJobResponse
    from dendro.common._api_request import _gui_delete_api_request
    resp = _gui_delete_api_request(url_path=f'/api/gui/jobs/{job_id}', github_access_token=github_access_token)
    resp = DeleteJobResponse(**resp)
    assert resp.success

def _delete_compute_resource(compute_resource_id: str, github_access_token: str):
    from dendro.api_helpers.routers.gui.compute_resource_routes import DeleteComputeResourceResponse
    from dendro.common._api_request import _gui_delete_api_request
    resp = _gui_delete_api_request(url_path=f'/api/gui/compute_resources/{compute_resource_id}', github_access_token=github_access_token)
    resp = DeleteComputeResourceResponse(**resp)
    assert resp.success

def _get_compute_resources(github_access_token: str):
    from dendro.api_helpers.routers.gui.compute_resource_routes import GetComputeResourcesResponse
    from dendro.common._api_request import _gui_get_api_request
    resp = _gui_get_api_request(url_path='/api/gui/compute_resources', github_access_token=github_access_token)
    resp = GetComputeResourcesResponse(**resp)
    assert resp.success
    return resp.computeResources

def _compute_resource_get_unfinished_jobs(compute_resource_id: str, compute_resource_private_key: str, wrong_payload: bool = False, wrong_signature: bool = False):
    from dendro.common.dendro_types import DendroJob
    from dendro.common._api_request import _compute_resource_get_api_request
    url_path = f'/api/compute_resource/compute_resources/{compute_resource_id}/unfinished_jobs'
    resp = _compute_resource_get_api_request(
        url_path=url_path,
        compute_resource_id=compute_resource_id,
        compute_resource_private_key=compute_resource_private_key,
        _wrong_payload_for_testing=wrong_payload,
        _wrong_signature_for_testing=wrong_signature
    )
    jobs = resp['jobs']
    jobs = [DendroJob(**job) for job in jobs]
    return jobs

def _get_job_output_upload_url(job_id: str, job_private_key: str, output_name: str):
    from dendro.api_helpers.routers.processor.router import ProcessorGetJobOutputUploadUrlResponse
    from dendro.common._api_request import _processor_get_api_request
    headers = {
        'job-private-key': job_private_key,
        'job-id': job_id
    }
    resp = _processor_get_api_request(url_path=f'/api/processor/jobs/{job_id}/outputs/{output_name}/upload_url', headers=headers)
    resp = ProcessorGetJobOutputUploadUrlResponse(**resp)
    return resp.uploadUrl

def _model_dump(model, exclude_none=False):
    # handle both pydantic v1 and v2
    if hasattr(model, 'model_dump'):
        return model.model_dump(exclude_none=exclude_none)
    else:
        return model.dict(exclude_none=exclude_none)



from fastapi import APIRouter
from .create_job_route import router as create_job_router
from .project_routes import router as project_router
from .compute_resource_routes import router as compute_resource_router
from .file_routes import router as file_router
from .job_routes import router as job_router
from .github_auth_routes import router as github_auth_router
from .user_routes import router as user_router
from .usage_routes import router as usage_router
from .find_routes import router as find_router

router = APIRouter()

router.include_router(create_job_router)
router.include_router(project_router, prefix="/projects")
router.include_router(compute_resource_router, prefix="/compute_resources")
router.include_router(file_router)
router.include_router(job_router, prefix="/jobs")
router.include_router(github_auth_router)
router.include_router(user_router, prefix="/users")
router.include_router(usage_router, prefix="/usage")
router.include_router(find_router)



from typing import Union, List
import os
import shutil
import requests
from ..mock import using_mock
from pydantic import BaseModel


class SetOutputFolderException(Exception):
    pass

class OutputFolder(BaseModel):
    name: Union[str, None] = None # the name of the output within the context of the processor (not needed when output_folder_name is specified for local testing)
    job_id: Union[str, None] = None
    job_private_key: Union[str, None] = None
    was_uploaded: bool = False
    output_folder_name: Union[str, None] = None # for local testing
    def upload(self, local_folder_name: str):
        from .Job import _get_upload_url_for_output_folder_file # avoid circular import

        if self.output_folder_name is not None:
            if self.job_id is not None:
                raise Exception('Cannot specify both output_folder_name and job_id in OutputFolder')
            if self.job_private_key is not None:
                raise Exception('Cannot specify both output_folder_name and job_private_key in OutputFolder')
            # copy local_folder_name to output_folder_name
            print(f'Copying output {self.name} to {self.output_folder_name}')
            shutil.copytree(local_folder_name, self.output_folder_name)
        else:
            if self.name is None:
                raise Exception('Unexpected: name is None in OutputFolder')
            if self.job_id is None:
                raise Exception('Unexpected: job_id is None in OutputFolder')
            if self.job_private_key is None:
                raise Exception('Unexpected: job_private_key is None in OutputFolder')

            all_relative_file_names = _get_all_relative_file_names(local_folder_name)
            if 'file_manifest.json' in all_relative_file_names:
                raise Exception('Cannot have a file named file_manifest.json in the output folder.')
            file_manifest = {
                'files': [
                    {
                        'name': relative_file_name,
                        'size': os.path.getsize(local_folder_name + '/' + relative_file_name)
                    }
                    for relative_file_name in all_relative_file_names
                ]
            }
            with open(local_folder_name + '/file_manifest.json', 'w') as f:
                import json
                json.dump(file_manifest, f)
            all_relative_file_names.append('file_manifest.json')
            for relative_file_name in all_relative_file_names:
                local_file_name = local_folder_name + '/' + relative_file_name
                upload_url = _get_upload_url_for_output_folder_file(name=self.name, relative_file_name=relative_file_name, job_id=self.job_id, job_private_key=self.job_private_key)

                # Upload the file to the URL
                print(f'Uploading output file {self.name} {relative_file_name}')

                with open(local_file_name, 'rb') as f:
                    if not using_mock():
                        resp_upload = requests.put(upload_url, data=f, timeout=60 * 60 * 24 * 7)
                        if resp_upload.status_code != 200:
                            print(upload_url)
                            raise SetOutputFolderException(f'Error uploading folder file to bucket ({resp_upload.status_code}) {resp_upload.reason}: {resp_upload.text}')

        self.was_uploaded = True

    # validator is needed to be an allowed pydantic type
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        else:
            raise ValueError(f'Unexpected type for OutputFolder: {type(value)}')

def _get_all_relative_file_names(local_folder_name: str):
    import os
    ret: List[str] = []
    for root, dirs, files in os.walk(local_folder_name):
        for file in files:
            relative_file_name = os.path.relpath(os.path.join(root, file), local_folder_name)
            ret.append(relative_file_name)
    return ret



from typing import Optional
from dendro.mock import using_mock
from ...core.settings import get_settings
from ._get_signed_upload_url import _get_signed_upload_url


# note that output_name of "_console_output" and "_resource_utilization_log" are special cases
async def get_upload_url(job_id: str, output_name: str):
    if output_name == "_console_output":
        pass
    elif output_name == "_resource_utilization_log":
        pass
    else:
        pass
        # not going to check this because it would require excessive database queries to get the job - there may be a way around it
        # aa = [x for x in job.outputFiles if x.name == output_name]
        # if len(aa) == 0:
        #     raise Exception(f"No output with name {output_name} **")

    object_key = f"dendro-outputs/{job_id}/{output_name}"

    upload_url, download_url = await _get_upload_url_for_object_key(object_key)
    return upload_url

async def get_upload_url_for_folder_file(job_id: str, output_folder_name: str, output_folder_file_name: str):
    object_key = f"dendro-outputs/{job_id}/{output_folder_name}/{output_folder_file_name}"

    upload_url, download_url = await _get_upload_url_for_object_key(object_key)
    return upload_url

async def get_additional_upload_url(*, job_id: str, sha1: str):
    if not _is_valid_sha1(sha1):
        raise Exception('Invalid sha1 string')

    object_key = f"dendro-outputs/{job_id}/sha1/{sha1}"

    upload_url, download_url = await _get_upload_url_for_object_key(object_key)
    return upload_url, download_url

async def _get_upload_url_for_object_key(object_key: str, size: Optional[int] = None):
    settings = get_settings()

    if using_mock():
        return f"https://mock-bucket.s3.amazonaws.com/{object_key}?mock-signature", f"https://mock-bucket/{object_key}"

    OUTPUT_BUCKET_URI = settings.OUTPUT_BUCKET_URI
    if OUTPUT_BUCKET_URI is None:
        raise Exception('Environment variable not set: OUTPUT_BUCKET_URI')
    OUTPUT_BUCKET_CREDENTIALS = settings.OUTPUT_BUCKET_CREDENTIALS
    if OUTPUT_BUCKET_CREDENTIALS is None:
        raise Exception('Environment variable not set: OUTPUT_BUCKET_CREDENTIALS')
    OUTPUT_BUCKET_BASE_URL = settings.OUTPUT_BUCKET_BASE_URL
    if OUTPUT_BUCKET_BASE_URL is None:
        raise Exception('Environment variable not set: OUTPUT_BUCKET_BASE_URL')
    if using_mock():
        output_bucket_base_url = 'https://mock-bucket'
    else:
        output_bucket_base_url = OUTPUT_BUCKET_BASE_URL

    signed_upload_url = await _get_signed_upload_url(
        bucket_uri=OUTPUT_BUCKET_URI,
        bucket_credentials=OUTPUT_BUCKET_CREDENTIALS,
        object_key=object_key,
        size=size
    )

    download_url = f'{output_bucket_base_url}/{object_key}'

    return signed_upload_url, download_url

def _is_valid_sha1(sha1: str):
    if len(sha1) != 40:
        return False
    for c in sha1:
        if c not in '0123456789abcdef':
            return False
    return True



from typing import Union
import boto3


def get_efs_fs_id(efs_fs_name: str):
    """
    Get the EFS file system ID from its name.
    References:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/efs.html#EFS.Client.describe_file_systems
    """
    try:
        client = boto3.client('efs')
        efs_fs_id = None
        response = client.describe_file_systems()
        for fs in response['FileSystems']:
            if fs.get("Name", "") == efs_fs_name:
                efs_fs_id = fs.get("FileSystemId")
        if efs_fs_id:
            return efs_fs_id
        else:
            raise Exception(f"EFS file system ID not found for name: {efs_fs_name}")
    except Exception as e:
        raise Exception(f"Error getting EFS file system ID: {e}")


def get_role_arn(role_name: str):
    """
    Get the role ARN from its name.
    References:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/iam/client/list_roles.html#list-roles
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/iam/client/list_role_tags.html#list-role-tags
    """
    try:
        iam_client = boto3.client('iam')
        required_tag_key = "DendroName"
        required_tag_value = role_name
        response = iam_client.list_roles()
        for role in response['Roles']:
            # Get the tags for each role
            role_name = role['RoleName']
            tags = iam_client.list_role_tags(RoleName=role_name)
            # Check if the role has the specific tag
            for tag in tags['Tags']:
                if tag['Key'] == required_tag_key and tag['Value'] == required_tag_value:
                    return role['Arn']
    except Exception as e:
        raise Exception(f"Error getting role ARN: {e}")


def create_aws_batch_job_definition(
    dendro_app_name: str,
    dendro_app_image_uri: str,
    job_role_name: str,
    efs_fs_name: Union[str, None] = None,
    environment_variables: Union[list, None] = None,
    container_command_override: Union[list, None] = None,
):
    """
    Create a job definition in AWS Batch.
    References:
    - https://boto3.amazonaws.com/v1/documentation/api/1.26.85/reference/services/batch/client/register_job_definition.html
    - https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task_definition_parameters.html#task_size
    - https://docs.aws.amazon.com/batch/latest/userguide/efs-volumes.html#specify-efs-config
    - https://aws.amazon.com/blogs/hpc/introducing-support-for-per-job-amazon-efs-volumes-in-aws-batch/
    """

    job_definition_name = f"{dendro_app_name}-dendro-jd"

    # Container Environment variables
    # e.g.: environment_variables = [{'name': 'string', 'value': 'string'}]
    if environment_variables is None:
        environment_variables = []
    else:
        for env_var in environment_variables:
            if not isinstance(env_var, dict):
                raise Exception("Environment variables must be a list of dictionaries.")
            if 'name' not in env_var or 'value' not in env_var:
                raise Exception("Environment variables must be a list of dictionaries with 'name' and 'value' keys. E.g.: [{'name': 'string', 'value': 'string'}]")

    # Container Resource requirements - these will always be provided in the submitted job
    resource_requirements = []

    # but the VCPU is required to be specified
    resource_requirements.append({
        'type': 'VCPU',
        'value': '1'
    })
    # and the MEMORY is required to be specified as well
    resource_requirements.append({
        'type': 'MEMORY',
        'value': str(1024 * 8)
    })

    # Get EFS file system ID and set container volumes and mount points
    volumes = []
    mount_points = []
    if efs_fs_name:
        efs_fs_id = get_efs_fs_id(efs_fs_name)
        if efs_fs_id:
            # Container Volumes and Mount Points configuration
            volumes = [
                {
                    'name': f'{job_definition_name}-EfsVolume',
                    'efsVolumeConfiguration': {
                        'fileSystemId': efs_fs_id,
                        'rootDirectory': "/",
                        # 'transitEncryption': 'ENABLED'|'DISABLED',
                        # 'transitEncryptionPort': 123,
                        # 'authorizationConfig': {
                        #     'accessPointId': 'string',
                        #     'iam': 'ENABLED'|'DISABLED'
                        # }
                    }
                },
            ]
            mount_points = [
                {
                    'containerPath': '/tmp',
                    'readOnly': False,
                    'sourceVolume': f'{job_definition_name}-EfsVolume'
                },
            ]

    # Get job role ARN
    job_role_arn = get_role_arn(job_role_name)

    # Container Command
    container_command = ["python", "/app/main.py"]
    if container_command_override:
        container_command = container_command_override

    # Container properties
    container_properties = {
        "image": dendro_app_image_uri,
        "command": container_command,
        "resourceRequirements": resource_requirements,
        "volumes": volumes,
        "mountPoints": mount_points,
        "environment": environment_variables,
        "jobRoleArn": job_role_arn,
    }

    try:
        client = boto3.client('batch')

        # deregister job definitions if they already exists
        response = client.describe_job_definitions(
            jobDefinitionName=job_definition_name,
            status='ACTIVE',
        )
        if len(response['jobDefinitions']) > 0:
            for job_definition in response['jobDefinitions']:
                print(f'Deregistering job definition: {job_definition["jobDefinitionArn"]}')
                client.deregister_job_definition(
                    jobDefinition=job_definition['jobDefinitionArn']
                )
        print(f'Registering job definition: {job_definition_name}')
        response = client.register_job_definition(
            jobDefinitionName=job_definition_name,
            type='container',
            retryStrategy={"attempts": 1},
            containerProperties=container_properties,
            timeout={"attemptDurationSeconds": 10800},
            platformCapabilities=['EC2'],
            tags={'CloudFormationStack': "DendroBatchStack"}
        )
        job_definition_arn = response['jobDefinitionArn']
        return job_definition_arn
    except Exception as e:
        raise Exception(f"Error creating job definition: {e}")



from ....common.dendro_types import ComputeResourceUserUsage
from ...clients.db import fetch_jobs_including_deleted


async def get_compute_resource_user_usage(*, compute_resource_id: str, user_id: str) -> ComputeResourceUserUsage:
    jobs_including_deleted = await fetch_jobs_including_deleted(compute_resource_id=compute_resource_id, user_id=user_id)
    return ComputeResourceUserUsage(
        computeResourceId=compute_resource_id,
        userId=user_id,
        jobsIncludingDeleted=jobs_including_deleted
    )



from typing import Union, List
import traceback
import aiohttp
from fastapi import APIRouter, HTTPException, Header
from .... import BaseModel
from ...services.processor.update_job_status import update_job_status
from ...services.processor.get_upload_url import get_upload_url, get_additional_upload_url, get_upload_url_for_folder_file
from ....common.dendro_types import ProcessorGetJobResponse, ProcessorGetJobResponseInput, ProcessorGetJobResponseOutput, ProcessorGetJobResponseOutputFolder, ProcessorGetJobResponseParameter, ProcessorGetJobResponseInputFolder, ProcessorGetJobResponseInputFolderFile
from ...clients.db import fetch_job, fetch_file, fetch_job_private_key

router = APIRouter()

# get job
@router.get("/jobs/{job_id}")
async def processor_get_job(job_id: str, job_private_key: str = Header(...)) -> ProcessorGetJobResponse:
    try:
        job = await fetch_job(job_id, include_dandi_api_key=True, include_secret_params=True, include_private_key=True)
        if job is None:
            raise Exception(f"No job with ID {job_id}")

        if job.jobPrivateKey != job_private_key:
            raise Exception(f"Invalid job private key for job {job_id}")
        # after we are done with it, let's delete the private key from the job object for safety
        job.jobPrivateKey = ''

        inputs: List[ProcessorGetJobResponseInput] = []
        input_folders: List[ProcessorGetJobResponseInputFolder] = []
        for input in job.inputFiles:
            file = await fetch_file(project_id=job.projectId, file_name=input.fileName)
            if file is None:
                raise Exception(f"Project file not found: {input.fileName}")
            if not file.content.startswith('url:'):
                raise Exception(f"Project file {input.fileName} is not a URL")
            url = file.content[len('url:'):]
            if not input.isFolder:
                # regular file
                if file.isFolder:
                    raise Exception(f"Mismatch. input is regular file but project file {input.fileName} is a folder")
                url = await _resolve_dandi_url(url, dandi_api_key=job.dandiApiKey)
                inputs.append(ProcessorGetJobResponseInput(
                    name=input.name,
                    url=url
                ))
            else:
                # folder
                if not file.isFolder:
                    raise Exception(f"Mismatch. input is folder but project file {input.fileName} is not a folder")
                file_manifest_obj = await _download_file_manifest_obj(url)
                # get all files in the folder
                folder_files: List[ProcessorGetJobResponseInputFolderFile] = []
                for f in file_manifest_obj['files']:
                    folder_files.append(ProcessorGetJobResponseInputFolderFile(
                        name=f['name'],
                        url=f'{url}/{f["name"]}',
                        size=f.get('size', None)
                    ))
                input_folders.append(ProcessorGetJobResponseInputFolder(
                    name=input.name,
                    files=folder_files
                ))

        outputs: List[ProcessorGetJobResponseOutput] = []
        output_folders: List[ProcessorGetJobResponseOutputFolder] = []
        for output in job.outputFiles:
            if not output.isFolder:
                # regular file
                outputs.append(ProcessorGetJobResponseOutput(
                    name=output.name
                ))
            else:
                # folder
                output_folders.append(ProcessorGetJobResponseOutputFolder(
                    name=output.name
                ))

        parameters: List[ProcessorGetJobResponseParameter] = []
        for parameter in job.inputParameters:
            parameters.append(ProcessorGetJobResponseParameter(
                name=parameter.name,
                value=parameter.value
            ))

        return ProcessorGetJobResponse(
            jobId=job.jobId,
            status=job.status,
            processorName=job.processorName,
            inputs=inputs,
            inputFolders=input_folders,
            outputs=outputs,
            outputFolders=output_folders,
            parameters=parameters
        )
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))

async def _resolve_dandi_url(url: str, *, dandi_api_key: Union[str, None]) -> str:
    if url.startswith('https://api.dandiarchive.org/api/') or url.startswith('https://api-staging.dandiarchive.org/api/'):
        headers = {}
        if dandi_api_key is not None:
            headers['Authorization'] = f'token {dandi_api_key}'
        async with aiohttp.ClientSession() as session:
            async with session.head(url, allow_redirects=True, headers=headers) as resp:
                return str(resp.url)
    else:
        return url

# update job status
class ProcessorUpdateJobStatusRequest(BaseModel):
    status: str
    error: Union[str, None] = None
    force_update: Union[bool, None] = None

class ProcessorUpdateJobStatusResponse(BaseModel):
    success: bool

@router.put("/jobs/{job_id}/status")
async def processor_update_job_status(job_id: str, data: ProcessorUpdateJobStatusRequest, job_private_key: str = Header(...)) -> ProcessorUpdateJobStatusResponse:
    try:
        job = await fetch_job(job_id, include_private_key=True)
        if job is None:
            raise Exception(f"No job with ID {job_id}")
        if job.jobPrivateKey != job_private_key:
            raise Exception(f"Invalid job private key for job {job_id}")

        await update_job_status(job=job, status=data.status, error=data.error, force_update=data.force_update)

        return ProcessorUpdateJobStatusResponse(success=True)
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))

# get job status
class ProcessorGetJobStatusResponse(BaseModel):
    status: Union[str, None] # return None if job does not exist
    success: bool

@router.get("/jobs/{job_id}/status")
async def processor_get_job_status(job_id: str, job_private_key: str = Header(...)) -> ProcessorGetJobStatusResponse:
    try:
        job = await fetch_job(job_id, include_private_key=True)
        if job is None:
            return ProcessorGetJobStatusResponse(status=None, success=True) # pragma: no cover
        if job.jobPrivateKey != job_private_key:
            raise Exception(f"Invalid job private key for job {job_id}") # pragma: no cover

        return ProcessorGetJobStatusResponse(status=job.status, success=True)
    except Exception as e: # pragma: no cover
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))

# get job output upload url
class ProcessorGetJobOutputUploadUrlResponse(BaseModel):
    uploadUrl: str
    success: bool

# note that output_name of "_console_output" and "_resource_utilization_log" are special cases
@router.get("/jobs/{job_id}/outputs/{output_name}/upload_url")
async def processor_get_upload_url(job_id: str, output_name: str, job_private_key: str = Header(...)) -> ProcessorGetJobOutputUploadUrlResponse:
    try:
        true_job_private_key = await fetch_job_private_key(job_id)
        assert true_job_private_key, f"No job with ID {job_id}"
        if true_job_private_key != job_private_key:
            raise ValueError(f"Invalid job private key for job {job_id}")

        signed_upload_url = await get_upload_url(job_id=job_id, output_name=output_name)
        return ProcessorGetJobOutputUploadUrlResponse(uploadUrl=signed_upload_url, success=True)
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e)) from e

# get job output folder file upload url
class ProcessorGetJobOutputFolderFileUploadUrlResponse(BaseModel):
    uploadUrl: str
    success: bool

@router.get("/jobs/{job_id}/output_folders/{output_folder_name}/files/{file_name:path}/upload_url")
async def processor_get_upload_url_for_output_folder_file(job_id: str, output_folder_name: str, file_name: str, job_private_key: str = Header(...)) -> ProcessorGetJobOutputFolderFileUploadUrlResponse:
    try:
        true_job_private_key = await fetch_job_private_key(job_id)
        assert true_job_private_key, f"No job with ID {job_id}"
        if true_job_private_key != job_private_key:
            raise ValueError(f"Invalid job private key for job {job_id}")

        signed_upload_url = await get_upload_url_for_folder_file(job_id=job_id, output_folder_name=output_folder_name, output_folder_file_name=file_name)
        return ProcessorGetJobOutputFolderFileUploadUrlResponse(uploadUrl=signed_upload_url, success=True)
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e)) from e

# get job output upload url for additional file
class ProcessorGetJobOutputUploadUrlForAdditionalFileResponse(BaseModel):
    uploadUrl: str
    downloadUrl: str
    success: bool

@router.get("/jobs/{job_id}/additional_uploads/sha1/{sha1}/upload_url")
async def processor_get_additional_upload_url(job_id: str, sha1: str, job_private_key: str = Header(...)) -> ProcessorGetJobOutputUploadUrlForAdditionalFileResponse:
    try:
        true_job_private_key = await fetch_job_private_key(job_id)
        assert true_job_private_key, f"No job with ID {job_id}"
        if true_job_private_key != job_private_key:
            raise ValueError(f"Invalid job private key for job {job_id}")

        signed_upload_url, download_url = await get_additional_upload_url(job_id=job_id, sha1=sha1)
        return ProcessorGetJobOutputUploadUrlForAdditionalFileResponse(uploadUrl=signed_upload_url, downloadUrl=download_url, success=True)
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e)) from e

async def _download_file_manifest_obj(folder_url: str):
    url = f'{folder_url}/file_manifest.json'
    print(f'Downloading file manifest from {url}')
    # first, we get the text content
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            if resp.status != 200:
                raise Exception(f"Error getting file manifest: {resp.status}")
            text = await resp.text()
    # then we parse it
    import json
    return json.loads(text)



import time
from typing import Union
from ...clients.db import fetch_file, delete_file, insert_file, update_project
from ....common.dendro_types import DendroFile
from ...core._create_random_id import _create_random_id
from .._remove_detached_files_and_jobs import _remove_detached_files_and_jobs


async def set_file(
    user_id: str,
    project_id: str,
    file_name: str,
    content: str, # for example, url:https://...
    job_id: Union[str, None],
    size: int,
    metadata: dict,
    is_folder: bool = False
):
    existing_file = await fetch_file(project_id, file_name)
    if existing_file is not None:
        await delete_file(project_id, file_name)
        deleted_old_file = True
    else:
        deleted_old_file = False

    new_file = DendroFile(
        projectId=project_id,
        fileId=_create_random_id(8),
        userId=user_id,
        fileName=file_name,
        size=size,
        timestampCreated=time.time(),
        content=content,
        metadata=metadata,
        isFolder=is_folder,
        jobId=job_id
    )
    await insert_file(new_file)

    if deleted_old_file:
        await _remove_detached_files_and_jobs(project_id)

    await update_project(
        project_id=project_id,
        update={
            'timestampModified': time.time()
        }
    )

    return new_file.fileId



from typing import Any, List

class ProcessorBase:
    name: str
    label: str
    description: str
    tags: List[str]
    attributes: dict

    @staticmethod
    def run(
        context: Any
    ):
        raise NotImplementedError()



from typing import Dict, Union, List
import uuid


class MockMongoClient:
    def __init__(self):
        self._dbs: Dict[str, MockMongoDatabase] = {}
    def __getitem__(self, key: str):
        if key not in self._dbs:
            self._dbs[key] = MockMongoDatabase()
        return self._dbs[key]
    def clear_databases(self):
        self._dbs = {}

class MockMongoDatabase:
    def __init__(self):
        self._collections: Dict[str, MockMongoCollection] = {}
    def __getitem__(self, key: str):
        if key not in self._collections:
            self._collections[key] = MockMongoCollection()
        return self._collections[key]

class MockMongoCollection:
    def __init__(self):
        self._documents: Dict[str, Dict] = {}
    def find(self, query: Dict):
        return MockMongoCursor(self._documents, query)
    async def find_one(self, query: Dict):
        for document in self._documents.values():
            if _document_matches_query(document, query):
                return document
        return None
    async def update_one(self, query: Dict, update: Dict, *, upsert=False):
        if '$set' not in update:
            raise NotImplementedError() # pragma: no cover
        update_val = update['$set']
        for document in self._documents.values():
            if _document_matches_query(document, query):
                document.update(update_val)
                return
        if upsert:
            # create a random ID
            update_val_2 = {
                **update_val,
                '_id': str(uuid.uuid4())
            }
            self._documents[update_val_2['_id']] = update_val_2
            return
        raise KeyError("No document matches query") # pragma: no cover
    async def insert_one(self, document: Dict):
        # create a random ID
        _id = str(uuid.uuid4())
        document2 = {
            **document,
            '_id': _id
        }
        self._documents[_id] = document2
    async def delete_one(self, query: Dict):
        for key, document in self._documents.items():
            if _document_matches_query(document, query):
                del self._documents[key]
                return
        raise KeyError("No document matches query") # pragma: no cover
    async def delete_many(self, query: Dict):
        document_items = list(self._documents.items()) # need to do it this way because we're deleting from the dict
        for key, document in document_items:
            if _document_matches_query(document, query):
                del self._documents[key]
        return

class MockMongoCursor:
    def __init__(self, documents: Dict[str, Dict], query: Dict):
        self._documents = documents
        self._query = query
    async def to_list(self, length: Union[int, None]) -> List[Dict]:
        assert length is None, 'non-None length is not supported for now'
        documents: List[Dict] = []
        for document in self._documents.values():
            if _document_matches_query(document, self._query):
                documents.append(document)
        return documents

def _document_matches_query(document: Dict, query: Dict) -> bool:
    # handle $in
    for key, value in query.items():
        if isinstance(value, dict):
            if '$in' in value:
                if key not in document:
                    return False # pragma: no cover
                if document[key] not in value['$in']:
                    return False
            elif '$ne' in value:
                if document.get(key, None) == value['$ne']:
                    return False
            else:
                raise NotImplementedError()
        else:
            if key not in document:
                return False # pragma: no cover
            if isinstance(document[key], list):
                if value not in document[key]:
                    return False
            else:
                if document[key] != value:
                    return False
    return True



from typing import Union
import os
import requests
import h5py
from pydantic import BaseModel
import remfile
from .FileManifest import FileManifest


class InputFolderDownloadError(Exception):
    pass


class InputFolder(BaseModel):
    name: Union[str, None] = None # the name of the input within the context of the processor (not needed when one of url or local_folder_name is specified directly)
    url: Union[str, None] = None
    local_folder_name: Union[str, None] = None
    job_id: Union[str, None] = None
    job_private_key: Union[str, None] = None
    _file_manifest: Union[FileManifest, None] = None

    def get_url(self, relative_file_name: str) -> str:
        if self.url is not None:
            if self.local_folder_name is not None:
                raise Exception('Cannot specify both url and local_folder_name in InputFolder')
            if self.job_id is not None:
                raise Exception('Cannot specify both url and job_id in InputFolder')
            if self.job_private_key is not None:
                raise Exception('Cannot specify both url and job_private_key in InputFolder')
            return self.url + '/' + relative_file_name
        elif self.local_folder_name is not None:
            if self.job_id is not None:
                raise Exception('Cannot specify both local_folder_name and job_id in InputFolder')
            if self.job_private_key is not None:
                raise Exception('Cannot specify both local_folder_name and job_private_key in InputFolder')
            raise Exception('Cannot get url for local_folder in InputFolder')
        else:
            if self.job_id is None:
                raise Exception('Unexpected: job_id is None')
            if self.job_private_key is None:
                raise Exception('Unexpected: job_private_key is None')
            if self.name is None:
                raise Exception('Unexpected: name is None in InputFolder')
            from .Job import _get_file_manifest_for_input_folder # avoid circular import
            if self._file_manifest is None:
                self._file_manifest = _get_file_manifest_for_input_folder(name=self.name, job_id=self.job_id, job_private_key=self.job_private_key)
            ff = next((f for f in self._file_manifest.files if f.name == relative_file_name), None)
            if ff is None:
                raise Exception(f'Input folder file not found when trying to get download URL: {self.name} {relative_file_name}')
            return ff.url

    def download(self, dest_folder_path: str):
        if self.local_folder_name is not None:
            # In the case of a local folder, we just copy it
            print(f'Copying {self.local_folder_name} to {dest_folder_path}')
            import shutil
            shutil.copytree(self.local_folder_name, dest_folder_path)
            return
        if self.job_id is None:
            raise Exception('Unexpected: job_id is None')
        if self.job_private_key is None:
            raise Exception('Unexpected: job_private_key is None')
        if self.name is None:
            raise Exception('Unexpected: name is None in InputFolder')
        from .Job import _get_file_manifest_for_input_folder # avoid circular import
        if self._file_manifest is None:
            self._file_manifest = _get_file_manifest_for_input_folder(name=self.name, job_id=self.job_id, job_private_key=self.job_private_key)
        os.mkdir(dest_folder_path)
        for ff in self._file_manifest.files:
            file_name = ff.name
            file_url = ff.url
            dest_file_path = os.path.join(dest_folder_path, file_name)
            # make sure the parent directory exists
            os.makedirs(os.path.dirname(dest_file_path), exist_ok=True)
            print(f'Downloading {file_url} to {dest_file_path}')
            # stream the download
            r = requests.get(file_url, stream=True, timeout=60 * 60 * 24 * 7)
            if r.status_code != 200:
                raise InputFolderDownloadError(f'Error downloading file {file_url}: {r.status_code} {r.reason}')
            with open(dest_file_path, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:
                        f.write(chunk)

    def get_file(self, relative_file_name, *, download: bool = False):
        if self.local_folder_name is not None:
            # In the case of a local file, we just return a file object
            f = open(os.path.join(self.local_folder_name, relative_file_name), 'rb')
            # An issue here is that this file is never closed. Not sure how to fix that.
            return f

        if download:
            if not os.path.exists('_download_inputs'):
                os.mkdir('_download_inputs')
            tmp_fname = f'_download_inputs/{self.name}'
            self.download(tmp_fname)
            f = open(os.path.join(tmp_fname, relative_file_name), 'rb')
            # An issue here is that this file is never closed. Not sure how to fix that.
            return f

        # We need to create an object that has a get_url() method
        # It's important to do it this way so that the url can renew as needed
        class FileInFolder:
            def __init__(self, input_folder: InputFolder, relative_file_name: str):
                self._input_folder = input_folder
                self._relative_file_name = relative_file_name

            def get_url(self):
                return self._input_folder.get_url(self._relative_file_name)
        the_file = FileInFolder(input_folder=self, relative_file_name=relative_file_name)
        return remfile.File(the_file)

    def get_h5py_file(self, relative_file_name):
        remf = self.get_file(relative_file_name)
        return h5py.File(remf, 'r')

    # validator is needed to be an allowed pydantic type
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        else:
            raise ValueError(f'Unexpected type for InputFolder: {type(value)}')



import os
from dendro.client import load_project


def test_load_project():
    if os.getenv('NO_INTERNET') == '1':
        print('Skipping test_load_project because NO_INTERNET is set to 1')
        return

    # if this project or file disappears then we'll need to update that here
    project_id = 'eb87e88a'
    file_name = 'imported/000618/sub-paired-english/sub-paired-english_ses-paired-english-m108-191125-163508_ecephys.nwb'
    project = load_project(project_id)
    file = project.get_file(file_name)
    assert file is not None
    assert file.get_url().startswith('https://')

    folder = project.get_folder('imported')
    assert len(folder.get_folders()) >= 1



# type: ignore

from typing import Union, List
import numpy as np
import h5py
import spikeinterface as si


class NwbRecording(si.BaseRecording):
    def __init__(self,
        file: h5py.File,
        electrical_series_path: str
    ) -> None:
        electrical_series: h5py.Group = file[electrical_series_path]
        electrical_series_data = electrical_series['data']
        dtype = electrical_series_data.dtype

        # Get sampling frequency
        if 'starting_time' in electrical_series.keys():
            # t_start = electrical_series['starting_time'][()]
            sampling_frequency = electrical_series['starting_time'].attrs['rate']
        elif 'timestamps' in electrical_series.keys():
            # t_start = electrical_series['timestamps'][0]
            sampling_frequency = 1 / np.median(np.diff(electrical_series['timestamps'][:1000]))

        # Get channel ids
        electrode_indices = electrical_series['electrodes'][:]
        electrodes_table = file['/general/extracellular_ephys/electrodes']
        channel_ids = [electrodes_table['id'][i] for i in electrode_indices]

        si.BaseRecording.__init__(self, channel_ids=channel_ids, sampling_frequency=sampling_frequency, dtype=dtype)

        # Set electrode locations
        if 'x' in electrodes_table:
            channel_loc_x = [electrodes_table['x'][i] for i in electrode_indices]
            channel_loc_y = [electrodes_table['y'][i] for i in electrode_indices]
            if 'z' in electrodes_table:
                channel_loc_z = [electrodes_table['z'][i] for i in electrode_indices]
            else:
                channel_loc_z = None
        elif 'rel_x' in electrodes_table:
            channel_loc_x = [electrodes_table['rel_x'][i] for i in electrode_indices]
            channel_loc_y = [electrodes_table['rel_y'][i] for i in electrode_indices]
            if 'rel_z' in electrodes_table:
                channel_loc_z = [electrodes_table['rel_z'][i] for i in electrode_indices]
            else:
                channel_loc_z = None
        else:
            channel_loc_x = None
            channel_loc_y = None
            channel_loc_z = None
        if channel_loc_x is not None:
            ndim = 2 if channel_loc_z is None else 3
            locations = np.zeros((len(electrode_indices), ndim), dtype=float)
            for i, electrode_index in enumerate(electrode_indices):
                locations[i, 0] = channel_loc_x[electrode_index]
                locations[i, 1] = channel_loc_y[electrode_index]
                if channel_loc_z is not None:
                    locations[i, 2] = channel_loc_z[electrode_index]
            self.set_dummy_probe_from_locations(locations)

        # Extractors channel groups must be integers, but Nwb electrodes group_name can be strings
        if "group_name" in electrodes_table:
            unique_electrode_group_names = list(np.unique(electrodes_table["group_name"][:]))

            groups = []
            for electrode_index in electrode_indices:
                group_name = electrodes_table["group_name"][electrode_index]
                group_id = unique_electrode_group_names.index(group_name)
                groups.append(group_id)
            self.set_channel_groups(groups)

        recording_segment = NwbRecordingSegment(
            electrical_series_data=electrical_series_data,
            sampling_frequency=sampling_frequency
        )
        self.add_recording_segment(recording_segment)

class NwbRecordingSegment(si.BaseRecordingSegment):
    def __init__(self, electrical_series_data: h5py.Dataset, sampling_frequency: float) -> None:
        self._electrical_series_data = electrical_series_data
        si.BaseRecordingSegment.__init__(self, sampling_frequency=sampling_frequency)

    def get_num_samples(self) -> int:
        return self._electrical_series_data.shape[0]

    def get_traces(self, start_frame: int, end_frame: int, channel_indices: Union[List[int], None] = None) -> np.ndarray:
        if channel_indices is None:
            return self._electrical_series_data[start_frame:end_frame, :]
        else:
            return self._electrical_series_data[start_frame:end_frame, channel_indices]



from fastapi import APIRouter, Header
from .... import BaseModel
from ...services._remove_detached_files_and_jobs import _remove_detached_files_and_jobs
from ....common.dendro_types import DendroJob
from ._authenticate_gui_request import _authenticate_gui_request
from ...core._get_project_role import _check_user_can_edit_project
from ...clients.db import fetch_compute_resource, fetch_job, fetch_project, delete_job as db_delete_job, approve_job as db_approve_job
from ...clients.pubsub import publish_pubsub_message
from ..common import api_route_wrapper


router = APIRouter()

# get job
class GetJobResponse(BaseModel):
    job: DendroJob
    success: bool

class JobNotFoundException(Exception):
    pass

@router.get("/{job_id}")
@api_route_wrapper
async def get_job(job_id) -> GetJobResponse:
    job = await fetch_job(job_id, raise_on_not_found=True)
    assert job
    return GetJobResponse(job=job, success=True)

# delete job
class DeleteJobResponse(BaseModel):
    success: bool

@router.delete("/{job_id}")
@api_route_wrapper
async def delete_job(job_id, github_access_token: str = Header(...)) -> DeleteJobResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    job = await fetch_job(job_id, raise_on_not_found=True)
    assert job

    project = await fetch_project(job.projectId, raise_on_not_found=True)
    assert project

    _check_user_can_edit_project(project, user_id)

    await db_delete_job(job_id)

    # remove detached files and jobs
    await _remove_detached_files_and_jobs(job.projectId)

    return DeleteJobResponse(success=True)

# approve job
class ApproveJobResponse(BaseModel):
    success: bool

@router.post("/{job_id}/approve")
@api_route_wrapper
async def approve_job(job_id, github_access_token: str = Header(...)) -> ApproveJobResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    job = await fetch_job(job_id, raise_on_not_found=True)
    assert job

    compute_resource_id = job.computeResourceId

    compute_resource = await fetch_compute_resource(compute_resource_id, raise_on_not_found=True)
    assert compute_resource

    if compute_resource.ownerId != user_id:
        raise Exception('Only the compute resource owner can approve jobs')

    await db_approve_job(job_id)

    # not really a status change, but we need to notify the compute resource
    await publish_pubsub_message(
        channel=job.computeResourceId,
        message={
            'type': 'jobStatusChanged',
            'projectId': job.projectId,
            'computeResourceId': job.computeResourceId,
            'jobId': job.jobId,
            'status': job.status
        }
    )

    return ApproveJobResponse(success=True)



from typing import Union
import time
import aiohttp
from ..clients._get_mongo_client import _get_mongo_client
from ..core._create_random_id import _create_random_id
from ._remove_detached_files_and_jobs import _remove_detached_files_and_jobs
from ...common.dendro_types import DendroFile
from ..core._model_dump import _model_dump
from ...mock import using_mock


async def _create_output_file(*,
    file_name: str,
    url: str,
    project_id: str,
    user_id: str,
    job_id: str,
    is_folder: Union[bool, None] = None,
    replace_pending: bool = False
) -> str: # returns the ID of the created file
    if url == 'pending':
        if replace_pending:
            raise Exception('Cannot replace pending file with another pending file')
        # this is the output of a job that has not completed yet
        size = 0
    else:
        if not using_mock():
            if not is_folder:
                size = await _get_size_for_remote_file(url)
            else:
                size = await _get_size_from_file_manifest_json(url)
        else:
            size = 1 # size of mock file

    client = _get_mongo_client()
    projects_collection = client['dendro']['projects']
    files_collection = client['dendro']['files']

    existing_file = await files_collection.find_one({
        'projectId': project_id,
        'fileName': file_name
    })
    existing_file = DendroFile(**existing_file) if existing_file is not None else None
    deleted_old_file = False
    if existing_file is not None:
        if replace_pending:
            if existing_file.content != 'pending':
                raise Exception('Error replacing pending file: existing file is not pending')
            file_id = existing_file.fileId # we will replace this file and it's important to use the same file ID
            await files_collection.delete_one({
                'projectId': project_id,
                'fileId': file_id
            })
        else:
            deleted_old_file = True
            file_id = _create_random_id(8)
            await files_collection.delete_one({
                'projectId': project_id,
                'fileName': file_name
            })
    else:
        if replace_pending:
            raise Exception('Cannot replace pending file because it does not exist')
        file_id = _create_random_id(8)

    new_file = DendroFile(
        projectId=project_id,
        fileId=file_id,
        userId=user_id,
        fileName=file_name,
        size=size,
        timestampCreated=time.time(),
        content=f'url:{url}' if url != 'pending' else 'pending',
        metadata={},
        isFolder=is_folder,
        jobId=job_id
    )
    await files_collection.insert_one(_model_dump(new_file, exclude_none=True))

    if deleted_old_file:
        await _remove_detached_files_and_jobs(project_id)

    await projects_collection.update_one({
        'projectId': project_id
    }, {
        '$set': {
            'timestampModified': time.time()
        }
    })

    return new_file.fileId

class GetSizeForRemoteFileException(Exception):
    pass

async def _get_size_for_remote_file(url: str) -> int:
    response = await _head_request(url)
    if response is None:
        raise GetSizeForRemoteFileException(f"Unable to HEAD {url}")
    size = response.headers.get('content-length')
    if size is None:
        raise GetSizeForRemoteFileException(f"Unable to get content-length for {url}")
    return int(size)

async def _head_request(url: str):
    async with aiohttp.ClientSession() as session:
        async with session.head(url) as response:
            return response

async def _get_size_from_file_manifest_json(url: str) -> int:
    file_manifest_url = url + '/file_manifest.json'
    try:
        # load the file manifest
        async with aiohttp.ClientSession() as session:
            async with session.get(file_manifest_url) as response:
                if response.status != 200:
                    raise Exception(f"Error getting file manifest: {response.status}")
                file_manifest = await response.json()
        size = 0
        for f in file_manifest['files']:
            size += f['size']
        return size
    except Exception as e:
        print(f'Problem reading file manifest at {file_manifest_url}: {e}')
        return 0



import time
from typing import List, Union
from ._get_mongo_client import _get_mongo_client
from ._remove_id_field import _remove_id_field
from ...common.dendro_types import DendroProject, DendroFile, DendroJob, DendroComputeResource, ComputeResourceSpec, DendroUser
from ..core._get_project_role import _project_has_user
from ..core._model_dump import _model_dump
from ..core._hide_secret_params_in_job import _hide_secret_params_in_job


async def fetch_projects_for_user(user_id: Union[str, None]) -> List[DendroProject]:
    client = _get_mongo_client()
    projects_collection = client['dendro']['projects']
    projects = await projects_collection.find({}).to_list(length=None) # type: ignore
    for project in projects:
        _remove_id_field(project)
    projects = [DendroProject(**project) for project in projects] # validate projects
    projects2: List[DendroProject] = []
    for project in projects:
        if _project_has_user(project, user_id):
            projects2.append(project)
    return projects2

async def fetch_all_projects() -> List[DendroProject]:
    client = _get_mongo_client()
    projects_collection = client['dendro']['projects']
    projects = await projects_collection.find({}).to_list(length=None) # type: ignore
    for project in projects:
        _remove_id_field(project)
    projects = [DendroProject(**project) for project in projects] # validate projects
    return projects

async def fetch_projects_with_tag(tag: str) -> List[DendroProject]:
    client = _get_mongo_client()
    projects_collection = client['dendro']['projects']
    projects = await projects_collection.find({
        # When you use a query like { "tags": tag } against an array field in MongoDB, it checks if any element of the array matches the value.
        'tags': tag
    }).to_list(length=None) # type: ignore
    for project in projects:
        _remove_id_field(project)
    projects = [DendroProject(**project) for project in projects] # validate projects
    return projects

async def fetch_project(project_id: str, *, raise_on_not_found=False) -> Union[DendroProject, None]:
    client = _get_mongo_client()
    projects_collection = client['dendro']['projects']
    project = await projects_collection.find_one({'projectId': project_id})
    # here I'd like to validate project
    _remove_id_field(project)
    if project is None:
        if raise_on_not_found:
            raise Exception(f"No project with ID {project_id}")
        else:
            return None
    return DendroProject(**project) # validate project

async def fetch_project_files(project_id: str, *, pending_only=False) -> List[DendroFile]:
    client = _get_mongo_client()
    files_collection = client['dendro']['files']
    query = {
        'projectId': project_id
    }
    if pending_only:
        query['content'] = 'pending'
    files = await files_collection.find(query).to_list(length=None) # type: ignore
    for file in files:
        _remove_id_field(file)
    files = [DendroFile(**file) for file in files] # validate files
    return files

async def fetch_multi_project_files(project_ids: List[str], *, pending_only=False) -> List[DendroFile]:
    client = _get_mongo_client()
    files_collection = client['dendro']['files']
    query: dict = {
        'projectId': {'$in': project_ids}
    }
    if pending_only:
        query['content'] = 'pending'
    files = await files_collection.find(query).to_list(length=None) # type: ignore
    for file in files:
        _remove_id_field(file)
    files = [DendroFile(**file) for file in files] # validate files
    return files

async def fetch_project_jobs(project_id: str, include_private_keys=False) -> List[DendroJob]:
    client = _get_mongo_client()
    jobs_collection = client['dendro']['jobs']
    jobs = await jobs_collection.find({
        'projectId': project_id,
        'deleted': {'$ne': True}
    }).to_list(length=None) # type: ignore
    for job in jobs:
        _remove_id_field(job)
    jobs = [DendroJob(**job) for job in jobs] # validate jobs
    if not include_private_keys:
        for job in jobs:
            job.jobPrivateKey = '' # hide the private key
    for job in jobs:
        job.dandiApiKey = None # hide the DANDI API key
        _hide_secret_params_in_job(job)
    return jobs

async def update_project(project_id: str, update: dict):
    client = _get_mongo_client()
    projects_collection = client['dendro']['projects']
    await projects_collection.update_one({
        'projectId': project_id
    }, {
        '$set': update
    })

async def delete_project(project_id: str):
    client = _get_mongo_client()
    projects_collection = client['dendro']['projects']
    await projects_collection.delete_one({
        'projectId': project_id
    })

async def delete_all_files_in_project(project_id: str):
    client = _get_mongo_client()
    files_collection = client['dendro']['files']
    await files_collection.delete_many({
        'projectId': project_id
    })

async def delete_all_jobs_in_project(project_id: str):
    client = _get_mongo_client()
    jobs_collection = client['dendro']['jobs']
    jobs = await jobs_collection.find({
        'projectId': project_id,
        'deleted': {'$ne': True}
    }).to_list(length=None) # type: ignore
    for job in jobs:
        jobs_collection.update_one({
            'jobId': job['jobId']
        }, {
            '$set': {
                'deleted': True
            }
        })

async def insert_project(project: DendroProject):
    client = _get_mongo_client()
    projects_collection = client['dendro']['projects']
    await projects_collection.insert_one(_model_dump(project, exclude_none=True))

async def fetch_compute_resource(compute_resource_id: str, raise_on_not_found=False):
    client = _get_mongo_client()
    compute_resources_collection = client['dendro']['computeResources']
    compute_resource = await compute_resources_collection.find_one({'computeResourceId': compute_resource_id})
    if compute_resource is None:
        if raise_on_not_found:
            raise ComputeResourceNotFoundError(f"No compute resource with ID {compute_resource_id}")
        else:
            return None # pragma: no cover (not ever run with raise_on_not_found=False)
    _remove_id_field(compute_resource)
    compute_resource = DendroComputeResource(**compute_resource) # validate compute resource
    return compute_resource

async def fetch_compute_resources_for_user(user_id: str):
    client = _get_mongo_client()
    compute_resources_collection = client['dendro']['computeResources']
    compute_resources = await compute_resources_collection.find({'ownerId': user_id}).to_list(length=None) # type: ignore
    for compute_resource in compute_resources:
        _remove_id_field(compute_resource)
    compute_resources = [DendroComputeResource(**compute_resource) for compute_resource in compute_resources] # validate compute resources
    return compute_resources

async def update_compute_resource(compute_resource_id: str, update: dict):
    client = _get_mongo_client()
    compute_resources_collection = client['dendro']['computeResources']
    await compute_resources_collection.update_one({
        'computeResourceId': compute_resource_id
    }, {
        '$set': update
    })

async def delete_compute_resource(compute_resource_id: str):
    client = _get_mongo_client()
    compute_resources_collection = client['dendro']['computeResources']
    await compute_resources_collection.delete_one({
        'computeResourceId': compute_resource_id
    })

async def register_compute_resource(compute_resource_id: str, name: str, user_id: str):
    client = _get_mongo_client()
    compute_resources_collection = client['dendro']['computeResources']

    compute_resource = await compute_resources_collection.find_one({'computeResourceId': compute_resource_id})
    if compute_resource is not None:
        await compute_resources_collection.update_one({'computeResourceId': compute_resource_id}, {
            '$set': {
                'ownerId': user_id,
                'name': name,
                'timestampModified': time.time()
            }
        })
    else:
        new_compute_resource = DendroComputeResource(
            computeResourceId=compute_resource_id,
            ownerId=user_id,
            name=name,
            timestampCreated=time.time(),
            apps=[]
        )
        await compute_resources_collection.insert_one(_model_dump(new_compute_resource, exclude_none=True))

async def fetch_compute_resource_jobs(compute_resource_id: str, statuses: Union[List[str], None], exclude_those_pending_approval: bool = False, include_private_keys: bool = False) -> List[DendroJob]:
    client = _get_mongo_client()
    jobs_collection = client['dendro']['jobs']
    query = {
        'computeResourceId': compute_resource_id,
        'deleted': {'$ne': True}
    }
    if statuses is not None:
        query['status'] = {'$in': statuses}
    if exclude_those_pending_approval:
        query['pendingApproval'] = {'$ne': True}
    jobs = await jobs_collection.find(query).to_list(length=None) # type: ignore
    for job in jobs:
        _remove_id_field(job)
    jobs = [DendroJob(**job) for job in jobs] # validate jobs
    if not include_private_keys:
        for job in jobs:
            job.jobPrivateKey = '' # hide the private key
    for job in jobs:
        job.dandiApiKey = None # hide the DANDI API key
        _hide_secret_params_in_job(job)
    return jobs

class ComputeResourceNotFoundError(Exception):
    pass

async def set_compute_resource_spec(compute_resource_id: str, spec: ComputeResourceSpec):
    client = _get_mongo_client()
    compute_resources_collection = client['dendro']['computeResources']
    compute_resource = await compute_resources_collection.find_one({'computeResourceId': compute_resource_id})
    assert compute_resource is not None, f"No compute resource with ID {compute_resource_id}"
    await compute_resources_collection.update_one({'computeResourceId': compute_resource_id}, {
        '$set': {
            'spec': _model_dump(spec, exclude_none=True)
        }
    })

class JobNotFoundError(Exception):
    pass

async def fetch_job(job_id: str, *, include_dandi_api_key: bool = False, include_secret_params: bool = False, include_private_key: bool = False, raise_on_not_found=False):
    client = _get_mongo_client()
    jobs_collection = client['dendro']['jobs']
    job = await jobs_collection.find_one({'jobId': job_id})
    _remove_id_field(job)
    if job is None:
        if raise_on_not_found:
            raise JobNotFoundError(f"No job with ID {job_id}")
        else:
            return None # pragma: no cover (not ever run with raise_on_not_found=False)
    job = DendroJob(**job) # validate job
    if not include_dandi_api_key:
        job.dandiApiKey = None
    if not include_secret_params:
        _hide_secret_params_in_job(job)
    if not include_private_key:
        job.jobPrivateKey = ''
    return job

job_private_key_cache: dict[str, str] = {}
# expire it every hour
job_private_key_cache_expiration_time = 60 * 60
job_private_key_cache_last_expiration_time = time.time()
def _expire_job_private_key_cache():
    global job_private_key_cache_last_expiration_time
    global job_private_key_cache
    if time.time() - job_private_key_cache_last_expiration_time > job_private_key_cache_expiration_time:
        job_private_key_cache = {}
        job_private_key_cache_last_expiration_time = time.time()

async def fetch_job_private_key(job_id: str):
    _expire_job_private_key_cache()
    if job_id in job_private_key_cache:
        return job_private_key_cache[job_id]
    job = await fetch_job(job_id, include_private_key=True)
    assert job is not None, f"No job with ID {job_id}"
    job_private_key_cache[job_id] = job.jobPrivateKey
    return job.jobPrivateKey

async def update_job(job_id: str, update: dict):
    client = _get_mongo_client()
    jobs_collection = client['dendro']['jobs']
    await jobs_collection.update_one({
        'jobId': job_id
    }, {
        '$set': update
    })

async def delete_job(job_id: str):
    client = _get_mongo_client()
    jobs_collection = client['dendro']['jobs']

    await jobs_collection.update_one({
        'jobId': job_id
    }, {
        '$set': {
            'deleted': True
        }
    })

async def approve_job(job_id: str):
    client = _get_mongo_client()
    jobs_collection = client['dendro']['jobs']

    await jobs_collection.update_one({
        'jobId': job_id
    }, {
        '$set': {
            'pendingApproval': False
        }
    })

async def insert_job(job: DendroJob):
    client = _get_mongo_client()
    jobs_collection = client['dendro']['jobs']
    await jobs_collection.insert_one(_model_dump(job, exclude_none=True))

async def fetch_file(project_id: str, file_name: str):
    client = _get_mongo_client()
    files_collection = client['dendro']['files']
    file = await files_collection.find_one({
        'projectId': project_id,
        'fileName': file_name
    })
    _remove_id_field(file)
    if file is None:
        return None
    file = DendroFile(**file) # validate file
    return file

async def delete_file(project_id: str, file_name: str):
    client = _get_mongo_client()
    files_collection = client['dendro']['files']
    await files_collection.delete_one({
        'projectId': project_id,
        'fileName': file_name
    })

async def insert_file(file: DendroFile):
    client = _get_mongo_client()
    files_collection = client['dendro']['files']
    await files_collection.insert_one(_model_dump(file, exclude_none=True))

class UserNotFoundError(Exception):
    pass

async def fetch_user_for_dendro_api_key(dendro_api_key: str):
    client = _get_mongo_client()
    users_collection = client['dendro']['users']
    user = await users_collection.find_one({
        'dendroApiKey': dendro_api_key
    })
    _remove_id_field(user)
    if user is None:
        raise UserNotFoundError(f"No user with dendro API key {dendro_api_key}")
    user = DendroUser(**user) # validate user
    return user

async def set_dendro_api_key_for_user(user_id: str, dendro_api_key: str):
    client = _get_mongo_client()
    users_collection = client['dendro']['users']
    await users_collection.update_one({
        'userId': user_id
    }, {
        '$set': {
            'dendroApiKey': dendro_api_key
        }
    }, upsert=True)

async def fetch_jobs_including_deleted(*, compute_resource_id: str, user_id: str, include_private_keys=False):
    client = _get_mongo_client()
    jobs_collection = client['dendro']['jobs']
    jobs = await jobs_collection.find({
        'computeResourceId': compute_resource_id,
        'userId': user_id
    }).to_list(length=None) # type: ignore
    for job in jobs:
        _remove_id_field(job)
    jobs = [DendroJob(**job) for job in jobs] # validate jobs
    if not include_private_keys:
        for job in jobs:
            job.jobPrivateKey = '' # hide the private key
    return jobs

async def fetch_files_with_content_string(content_string: str) -> List[DendroFile]:
    client = _get_mongo_client()
    files_collection = client['dendro']['files']
    files = await files_collection.find({
        'content': content_string
    }).to_list(length=None)
    for file in files:
        _remove_id_field(file)
    files = [DendroFile(**file) for file in files] # validate files
    return files



import sys
import os
import time
from typing import Union, Optional, Dict, Any
import subprocess
from ..common._api_request import _processor_put_api_request
from ..mock import using_mock
from ..internal_job_monitoring.console_output_monitor import do_upload as _upload_final_console_output
import shutil


# This function is called internally by the compute resource daemon through the dendro CLI
# * Sets the job status to running in the database via the API
# * Runs the job in a separate process by calling the app executable with the appropriate env vars
# * Launches detached processes to monitor the console output, resource utilization, and job status
# * Finally, sets the job status to completed or failed in the database via the API

dendro_internal_folder = '_dendro'

def _run_job_parent_process(*, job_id: str, job_private_key: str, app_executable: str, job_timeout_sec: Union[int, None]):
    timescale = 1 if not using_mock() else 1000 # if using mock, run 1000x faster so that we can test timeouts

    _run_job_timer = time.time()

    if os.path.exists(dendro_internal_folder):
        shutil.rmtree(dendro_internal_folder)
    os.mkdir(dendro_internal_folder)
    console_out_fname = os.path.join(dendro_internal_folder, 'console_output.txt')
    cancel_out_fname = os.path.join(dendro_internal_folder, 'cancel.txt')
    console_out_monitor_output_fname = os.path.join(dendro_internal_folder, 'console_output_monitor_output.txt')
    resource_utilization_monitor_output_fname = os.path.join(dendro_internal_folder, 'resource_utilization_monitor_output.txt')
    job_status_monitor_output_fname = os.path.join(dendro_internal_folder, 'job_status_monitor_output.txt')

    # set the job status to running by calling the remote dendro API
    _debug_log(f'Running job {job_id}')
    _set_job_status(job_id=job_id, job_private_key=job_private_key, status='running')

    last_report_timestamp = 0

    proc = None

    with open(console_out_fname, 'w') as console_out_file, open(console_out_monitor_output_fname, 'w') as console_out_monitor_output_file, open(resource_utilization_monitor_output_fname, 'w') as resource_utilization_monitor_output_file, open(job_status_monitor_output_fname, 'w') as job_status_monitor_output_file:
        succeeded = False # whether we succeeded in running the job without an exception
        error_message = '' # if we fail, this will be set to the exception message
        try:
            if not using_mock():
                # start the console output monitor
                cmd = f'dendro internal-job-monitor console_output --parent-pid {os.getpid()}'
                env = {
                    'JOB_ID': job_id,
                    'JOB_PRIVATE_KEY': job_private_key,
                    'CONSOLE_OUT_FILE': os.path.abspath(console_out_fname)
                }
                _launch_detached_process(cmd=cmd, env=env, stdout=console_out_monitor_output_file, stderr=subprocess.STDOUT)

                # start the resource utilization monitor
                cmd = f'dendro internal-job-monitor resource_utilization --parent-pid {os.getpid()}'
                env = {
                    'JOB_ID': job_id,
                    'JOB_PRIVATE_KEY': job_private_key
                }
                _launch_detached_process(cmd=cmd, env=env, stdout=resource_utilization_monitor_output_file, stderr=subprocess.STDOUT)

                # start the status check monitor
                cmd = f'dendro internal-job-monitor job_status --parent-pid {os.getpid()}'
                env = {
                    'JOB_ID': job_id,
                    'JOB_PRIVATE_KEY': job_private_key,
                    'CANCEL_OUT_FILE': cancel_out_fname
                }
                _launch_detached_process(cmd=cmd, env=env, stdout=job_status_monitor_output_file, stderr=subprocess.STDOUT)

            # Launch the job in a separate process
            proc = _launch_job_child_process(
                job_id=job_id,
                job_private_key=job_private_key,
                app_executable=app_executable,
                console_out_file=console_out_file
            )

            first_iteration = True
            while True:
                skip_this_check = using_mock() and first_iteration
                if not skip_this_check:
                    try:
                        retcode = proc.wait(1 / timescale)
                        # don't check this now -- wait until after we had a chance to read the last console output
                    except subprocess.TimeoutExpired:
                        retcode = None # process is still running
                else:
                    retcode = None

                if retcode is not None:
                    if retcode != 0:
                        raise ValueError(f'Error running job: return code {retcode}')
                    # job has completed with exit code 0
                    break

                # check if the cancel file has been created
                if os.path.exists(cancel_out_fname):
                    with open(cancel_out_fname, 'r') as f:
                        cancel_msg = f.read()
                    _debug_log(f'Job canceled: {cancel_msg}')
                    raise Exception(f'Job canceled: {cancel_msg}')

                # check job timeout
                if job_timeout_sec is not None:
                    elapsed = time.time() - _run_job_timer
                    if elapsed > job_timeout_sec:
                        raise Exception(f'Job timed out: {elapsed} > {job_timeout_sec} seconds')

                first_iteration = False

                elapsed_since_report = time.time() - last_report_timestamp
                if elapsed_since_report >= 120:
                    last_report_timestamp = time.time()
                    _debug_log('Job still running')

                time.sleep(3 / timescale)
            succeeded = True # No exception
        except Exception as e: # pylint: disable=broad-except
            _debug_log(f'Error running job: {str(e)}')
            succeeded = False
            error_message = str(e)
        finally:
            if proc is not None:
                _debug_log('Closing subprocess')
                try:
                    if proc.stdout:
                        proc.stdout.close()
                    if proc.stderr:
                        proc.stderr.close()
                    proc.terminate()
                except Exception: # pylint: disable=broad-except
                    pass

            dendro_job_cleanup_dir = os.environ.get('DENDRO_JOB_CLEANUP_DIR', None)
            if dendro_job_cleanup_dir is not None:
                _debug_log(f'Cleaning up DENDRO_JOB_CLEANUP_DIR: {dendro_job_cleanup_dir}')
                try:
                    # delete files in the cleanup dir but do not delete the cleanup dir itself
                    def _delete_files_in_dir(dir: str):
                        for fname in os.listdir(dir):
                            fpath = os.path.join(dir, fname)
                            if os.path.isdir(fpath):
                                if fname == '_dendro':
                                    # don't delete the internal log folder
                                    continue
                                _delete_files_in_dir(fpath)
                            else:
                                _debug_log(f'Deleting {fpath}')
                                os.remove(fpath)
                    _delete_files_in_dir(dendro_job_cleanup_dir)
                except Exception as e:
                    _debug_log(f'WARNING: problem cleaning up DENDRO_JOB_CLEANUP_DIR: {str(e)}')
            else:
                _debug_log('No DENDRO_JOB_CLEANUP_DIR environment variable set. Not cleaning up.')

    if not using_mock():
        _debug_log('Uploading final console output')
        # this is needed because the console output monitor may get terminated before it has a chance to upload the final console output
        ok = _upload_final_console_output(
            job_id=job_id,
            job_private_key=job_private_key,
            console_out_file=console_out_fname
        )
        if not ok:
            _debug_log('WARNING: problem uploading final console output')

    # Set the final job status
    _debug_log('Finalizing job')
    _finalize_job(job_id=job_id, job_private_key=job_private_key, succeeded=succeeded, error_message=error_message)

    _debug_log('Exiting')

def _launch_job_child_process(*, job_id: str, job_private_key: str, app_executable: str, console_out_file: Any):
    if not using_mock(): # pragma: no cover
        # Set the appropriate environment variables and launch the job in a background process
        cmd = app_executable
        env = os.environ.copy()
        env = {
            **env,
            'JOB_ID': job_id,
            'JOB_PRIVATE_KEY': job_private_key,
            'JOB_INTERNAL': '1',
            'PYTHONUNBUFFERED': '1'
        }
        _debug_log(f'Running {app_executable} (Job ID: {job_id})) (Job private key: {job_private_key})')
        working_dir = os.environ.get('DENDRO_JOB_WORKING_DIR', None)
        if working_dir is not None:
            if not os.path.exists(working_dir):
                # make directory including parent directories
                os.makedirs(working_dir)
            if not os.path.isdir(working_dir + '/tmp'):
                os.mkdir(working_dir + '/tmp')
            env['DENDRO_JOB_WORKING_DIR'] = working_dir
            env['TMPDIR'] = working_dir + '/tmp'
            _debug_log(f'Using working directory {working_dir}')
        _debug_log('Opening subprocess')
        proc = subprocess.Popen(
            cmd,
            env=env,
            stdout=console_out_file,
            stderr=subprocess.STDOUT,
            cwd=working_dir
        )
        return proc
    else:
        # it's not going to work to run the mock job in a separate proecess because the mock environment will not exist
        # so instead we are going to run a dummy mock jub, and then for test coverage,
        # we are also going to do execute _run_job on the app
        proc = subprocess.Popen(
            ['dendro', 'run-mock-job'],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )

        # here's where we execute _run_job for test coverage
        from ..compute_resource._start_job import _load_app_from_main
        old_working_dir = os.getcwd()
        os.chdir(os.path.dirname(app_executable))
        app_instance = _load_app_from_main(app_executable)
        try:
            from ._run_job_child_process import _run_job_child_process
            _run_job_child_process(job_id=job_id, job_private_key=job_private_key, processors=app_instance._processors)

            return proc
        finally:
            os.chdir(old_working_dir)
            if 'main' in sys.modules:
                del sys.modules['main'] # important to do this so that at a later time we can load a different main.py

def _finalize_job(*, job_id: str, job_private_key: str, succeeded: bool, error_message: str):
    try:
        if succeeded:
            # The job has completed successfully - update the status accordingly
            _debug_log('Setting job status to completed')
            print('Job completed')
            _set_job_status(job_id=job_id, job_private_key=job_private_key, status='completed')
        else:
            # The job has failed - update the status accordingly and set the error message
            _debug_log('Setting job status to failed: ' + error_message)
            print('Job failed: ' + error_message)
            _set_job_status(job_id=job_id, job_private_key=job_private_key, status='failed', error=error_message)
    except Exception as e: # pylint: disable=broad-except
        # This is unfortunate - we completed the job, but somehow failed to update the status in the dendro system - maybe there was a network error (maybe we should retry?)
        _debug_log('WARNING: problem setting final job status: ' + str(e))
        print('WARNING: problem setting final job status: ' + str(e))

class SetJobStatusError(Exception):
    pass

def _set_job_status(*, job_id: str, job_private_key: str, status: str, error: Optional[str] = None):
    """Set the status of a job in the dendro API"""
    url_path = f'/api/processor/jobs/{job_id}/status'
    headers = {
        'job-private-key': job_private_key
    }
    data: Dict[str, Any] = {
        'status': status
    }
    if error is not None:
        data['error'] = error
    if os.environ.get('DENDRO_FORCE_STATUS_UPDATES', None) == '1':
        data['force_update'] = True
    resp = _processor_put_api_request(
        url_path=url_path,
        headers=headers,
        data=data
    )
    if not resp['success']:
        raise SetJobStatusError(f'Error setting job status: {resp["error"]}')

def _launch_detached_process(*, cmd: str, env: Dict[str, str], stdout: Any, stderr: Any):
    _debug_log(f'Launching detached process: {cmd}')
    subprocess.Popen(
        cmd.split(' '),
        env={
            **os.environ.copy(),
            **env
        },
        stdout=stdout,
        stderr=stderr,
        start_new_session=True
    )

def _debug_log(msg: str):
    timestamp = time.strftime('%Y-%m-%d %H:%M:%S')
    msg2 = f'{timestamp} {msg}'
    print(msg2)
    # write to dendro-job.log
    # this will be written to the working directory, which should be in the job dir
    with open(f'{dendro_internal_folder}/dendro-job.log', 'a', encoding='utf-8') as f:
        f.write(msg2 + '\n')



from typing import Any, List, Dict
import time
from dataclasses import dataclass
from .InputFile import InputFile
from .InputFolder import InputFolder
from .OutputFile import OutputFile
from .OutputFolder import OutputFolder
from ..common._api_request import _processor_get_api_request
from ..common.dendro_types import ProcessorGetJobResponse
from .FileManifest import FileManifest, FileManifestFile


@dataclass
class JobParameter:
    """A parameter passed to a job"""
    name: str
    value: Any

class JobApiRequestException(Exception):
    pass

class Job:
    """A job that is passed to a processor"""
    def __init__(self,
        job_id: str,
        job_private_key: str
    ) -> None:
        self._job_id = job_id
        self._job_private_key = job_private_key
        resp = _job_info_manager.get_job_info(job_id=job_id, job_private_key=job_private_key)
        # important to set these only once here because these objects will be passed into the processor function
        self._inputs = [InputFile(name=i.name, job_id=self._job_id, job_private_key=self._job_private_key) for i in resp.inputs]
        self._input_folders = [InputFolder(name=i.name, job_id=self._job_id, job_private_key=self._job_private_key) for i in resp.inputFolders] if resp.inputFolders else None
        self._outputs = [OutputFile(name=o.name, job_id=self._job_id, job_private_key=self._job_private_key) for o in resp.outputs]
        self._output_folders = [OutputFolder(name=o.name, job_id=self._job_id, job_private_key=self._job_private_key) for o in resp.outputFolders] if resp.outputFolders else None
        self._parameters = [JobParameter(name=p.name, value=p.value) for p in resp.parameters]
        self._processor_name = resp.processorName
    @property
    def job_id(self) -> str:
        """The ID of the job"""
        return self._job_id
    @property
    def processor_name(self) -> str:
        """The name of the processor"""
        resp = _job_info_manager.get_job_info(job_id=self._job_id, job_private_key=self._job_private_key)
        return resp.processorName
    @property
    def inputs(self) -> List[InputFile]:
        """The input files of the job"""
        return self._inputs
    @property
    def input_folders(self) -> List[InputFolder]:
        """The input folders of the job"""
        return self._input_folders if self._input_folders is not None else []
    @property
    def outputs(self) -> List[OutputFile]:
        """The output files of the job"""
        return self._outputs
    @property
    def output_folders(self) -> List[OutputFolder]:
        """The output folders of the job"""
        return self._output_folders if self._output_folders is not None else []
    @property
    def parameters(self) -> List[JobParameter]:
        """The parameters of the job"""
        return self._parameters

def _get_upload_url_for_output_file(*, name: str, job_id: str, job_private_key: str) -> str:
    """Get a signed upload URL for an output file"""

    url_path = f'/api/processor/jobs/{job_id}/outputs/{name}/upload_url'
    headers = {
        'job-private-key': job_private_key
    }
    resp = _processor_get_api_request(
        url_path=url_path,
        headers=headers
    )
    upload_url = resp['uploadUrl'] # This will be a presigned AWS S3 URL
    return upload_url

def _get_upload_url_for_output_folder_file(*, name: str, relative_file_name: str, job_id: str, job_private_key: str) -> str:
    """Get a signed upload URL for an output folder file"""

    url_path = f'/api/processor/jobs/{job_id}/output_folders/{name}/files/{relative_file_name}/upload_url'
    headers = {
        'job-private-key': job_private_key
    }
    resp = _processor_get_api_request(
        url_path=url_path,
        headers=headers
    )
    upload_url = resp['uploadUrl'] # This will be a presigned AWS S3 URL
    return upload_url

@dataclass
class JobInfoRecord:
    job_id: str
    job_private_key: str
    resp: ProcessorGetJobResponse
    timestamp: float

class JobInfoManager:
    def __init__(self) -> None:
        self._info_records: Dict[str, JobInfoRecord] = {}
    def get_job_info(self, *, job_id: str, job_private_key: str) -> ProcessorGetJobResponse:
        if job_id in self._info_records:
            rec = self._info_records[job_id]
            elapsed = time.time() - rec.timestamp
            if elapsed < 30 * 60:
                return rec.resp
        url_path = f'/api/processor/jobs/{job_id}'
        headers = {
            'job-private-key': job_private_key
        }
        resp_dict = _processor_get_api_request(
            url_path=url_path,
            headers=headers
        )
        resp = ProcessorGetJobResponse(**resp_dict)
        rec = JobInfoRecord(job_id=job_id, job_private_key=job_private_key, resp=resp, timestamp=time.time())
        self._info_records[job_id] = rec
        return resp

_job_info_manager = JobInfoManager()

def _get_download_url_for_input_file(*, name: str, job_id: str, job_private_key: str) -> str:
    resp = _job_info_manager.get_job_info(job_id=job_id, job_private_key=job_private_key)
    input = next((i for i in resp.inputs if i.name == name), None)
    if input is None:
        raise Exception(f'Input not found when trying to get download URL: {name}')
    download_url = input.url
    return download_url

def _get_file_manifest_for_input_folder(*, name: str, job_id: str, job_private_key: str) -> FileManifest:
    resp = _job_info_manager.get_job_info(job_id=job_id, job_private_key=job_private_key)
    if resp.inputFolders is None:
        raise Exception(f'Input folders not found when trying to get file manifest: {name}')
    input_folder = next((i for i in resp.inputFolders if i.name == name), None)
    if input_folder is None:
        raise Exception(f'Input folder not found when trying to get file manifest: {name}')
    manifest = FileManifest(
        files=[
            FileManifestFile(
                name=f.name,
                url=f.url,
                size=f.size
            )
            for f in input_folder.files
        ]
    )
    return manifest



from typing import List
import time

from ...clients.pubsub import publish_pubsub_message
from .... import BaseModel
from fastapi import APIRouter, Header
from ...services._crypto_keys import _verify_signature
from ....common.dendro_types import DendroComputeResource, DendroComputeResourceApp, DendroJob, PubsubSubscription
from ._authenticate_gui_request import _authenticate_gui_request
from ...clients.db import fetch_compute_resource, fetch_compute_resources_for_user, update_compute_resource, fetch_compute_resource_jobs
from ...clients.db import register_compute_resource as db_register_compute_resource
from ...clients.db import delete_compute_resource as db_delete_compute_resource
from ...core.settings import get_settings
from ...core._model_dump import _model_dump
from ....mock import using_mock
from ..common import api_route_wrapper, AuthException


router = APIRouter()

# get compute resource
class GetComputeResourceResponse(BaseModel):
    computeResource: DendroComputeResource
    success: bool

class ComputeResourceNotFoundException(Exception):
    pass

@router.get("/{compute_resource_id}")
@api_route_wrapper
async def get_compute_resource(compute_resource_id) -> GetComputeResourceResponse:
    compute_resource = await fetch_compute_resource(compute_resource_id, raise_on_not_found=True)
    assert compute_resource
    return GetComputeResourceResponse(computeResource=compute_resource, success=True)

# get compute resources
class GetComputeResourcesResponse(BaseModel):
    computeResources: List[DendroComputeResource]
    success: bool

@router.get("")
@api_route_wrapper
async def get_compute_resources(github_access_token: str = Header(...)):
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    compute_resources = await fetch_compute_resources_for_user(user_id)

    return GetComputeResourcesResponse(computeResources=compute_resources, success=True)

# set compute resource apps
class SetComputeResourceAppsRequest(BaseModel):
    apps: List[DendroComputeResourceApp]

class SetComputeResourceAppsResponse(BaseModel):
    success: bool

@router.put("/{compute_resource_id}/apps")
@api_route_wrapper
async def set_compute_resource_apps(compute_resource_id, data: SetComputeResourceAppsRequest, github_access_token: str = Header(...)) -> SetComputeResourceAppsResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    # parse the request
    apps = data.apps

    compute_resource = await fetch_compute_resource(compute_resource_id, raise_on_not_found=True)
    assert compute_resource

    if compute_resource.ownerId != user_id:
        raise AuthException('User does not have permission to admin this compute resource')

    await update_compute_resource(compute_resource_id, update={
        'apps': [_model_dump(app, exclude_none=True) for app in apps],
        'timestampModified': time.time()
    })

    await publish_pubsub_message(
        channel=compute_resource_id,
        message={
            'type': 'computeResourceAppsChanaged',
            'computeResourceId': compute_resource_id
        }
    )

    return SetComputeResourceAppsResponse(success=True)

# delete compute resource
class DeleteComputeResourceResponse(BaseModel):
    success: bool

@router.delete("/{compute_resource_id}")
@api_route_wrapper
async def delete_compute_resource(compute_resource_id, github_access_token: str = Header(...)) -> DeleteComputeResourceResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    compute_resource = await fetch_compute_resource(compute_resource_id, raise_on_not_found=True)
    assert compute_resource
    if compute_resource.ownerId != user_id:
        raise AuthException('User does not have permission to delete this compute resource')

    await db_delete_compute_resource(compute_resource_id)

    return DeleteComputeResourceResponse(success=True)

# get pubsub subscription
class GetPubsubSubscriptionResponse(BaseModel):
    subscription: PubsubSubscription
    success: bool

@router.get("/{compute_resource_id}/pubsub_subscription")
@api_route_wrapper
async def get_pubsub_subscription(compute_resource_id):
    compute_resource = await fetch_compute_resource(compute_resource_id, raise_on_not_found=True)
    assert compute_resource

    if using_mock():
        subscription = PubsubSubscription(
            pubnubSubscribeKey='mock-subscribe-key',
            pubnubChannel=compute_resource_id,
            pubnubUser=compute_resource_id
        )
    else: # pragma: no cover
        VITE_PUBNUB_SUBSCRIBE_KEY = get_settings().PUBNUB_SUBSCRIBE_KEY
        if VITE_PUBNUB_SUBSCRIBE_KEY is None:
            raise KeyError('Environment variable not set: VITE_PUBNUB_SUBSCRIBE_KEY (gui)')
        subscription = PubsubSubscription(
            pubnubSubscribeKey=VITE_PUBNUB_SUBSCRIBE_KEY,
            pubnubChannel=compute_resource_id,
            pubnubUser=compute_resource_id
        )
    return GetPubsubSubscriptionResponse(subscription=subscription, success=True)

# register compute resource
class RegisterComputeResourceRequest(BaseModel):
    computeResourceId: str
    resourceCode: str
    name: str

class RegisterComputeResourceResponse(BaseModel):
    success: bool

@router.post("/register")
@api_route_wrapper
async def register_compute_resource(data: RegisterComputeResourceRequest, github_access_token: str = Header(...)) -> RegisterComputeResourceResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    # parse the request
    compute_resource_id = data.computeResourceId
    resource_code = data.resourceCode
    name = data.name

    ok = _verify_resource_code(compute_resource_id, resource_code)
    if not ok:
        raise AuthException('Invalid resource code')

    await db_register_compute_resource(compute_resource_id=compute_resource_id, name=name, user_id=user_id)

    return RegisterComputeResourceResponse(success=True)

# get jobs for compute resource
class GetJobsForComputeResourceResponse(BaseModel):
    jobs: List[DendroJob]
    success: bool

@router.get("/{compute_resource_id}/jobs")
@api_route_wrapper
async def get_jobs_for_compute_resource(compute_resource_id, github_access_token: str = Header(...)) -> GetJobsForComputeResourceResponse:
    # authenticate the request
    user_id = await _authenticate_gui_request(github_access_token=github_access_token, raise_on_not_authenticated=True)
    assert user_id

    compute_resource = await fetch_compute_resource(compute_resource_id, raise_on_not_found=True)
    assert compute_resource
    if compute_resource.ownerId != user_id:
        raise AuthException('User does not have permission to view jobs for this compute resource')

    jobs = await fetch_compute_resource_jobs(compute_resource_id, statuses=None, include_private_keys=False)

    return GetJobsForComputeResourceResponse(jobs=jobs, success=True)

def _verify_resource_code(compute_resource_id: str, resource_code: str) -> bool:
    # check that timestamp is within 5 minutes of current time
    try:
        timestamp = int(resource_code.split('-')[0])
        if abs(timestamp - time.time()) > 300:
            return False
        signature = resource_code.split('-')[1]
        if not _verify_signature({'timestamp': timestamp}, compute_resource_id, signature):
            return False
        return True
    except Exception:
        print(f'Error verifying resource code: {resource_code}')
        return False



import os

from pydantic import VERSION as pydantic_version # noqa

# In the future we may want to intercept and override the pydantic BaseModel
from pydantic import BaseModel, Field # noqa

# read the version from thisdir/version.txt
thisdir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(thisdir, 'version.txt')) as f:
    __version__ = f.read().strip()



import time
from typing import Union

from dendro.mock import using_mock
from ....common.dendro_types import DendroJob
from ...core.settings import get_settings
from ...core._model_dump import _model_dump
from .._create_output_file import _create_output_file
from ...clients.db import update_job
from ...clients.pubsub import publish_pubsub_message


async def update_job_status(job: DendroJob, status: str, error: Union[str, None], force_update: Union[bool, None] = None):
    old_status = job.status

    new_status = status
    new_error = error

    if not force_update:
        if new_status == 'starting':
            if old_status != 'pending':
                raise Exception(f"Cannot set job status to starting when status is {old_status}")
        elif new_status == 'running':
            if old_status != 'starting':
                raise Exception(f"Cannot set job status to running when status is {old_status}")
        elif new_status == 'completed':
            if old_status != 'running':
                raise Exception(f"Cannot set job status to completed when status is {old_status}")
        elif new_status == 'failed':
            if (old_status != 'running') and (old_status != 'starting') and (old_status != 'pending'):
                raise Exception(f"Cannot set job status to failed when status is {old_status}")

    update = {}
    if new_error:
        if new_status != 'failed':
            raise Exception(f"Cannot set job error when status is {new_status}")
    if new_status == 'completed':
        # we need to create the output files before marking the job as completed
        output_bucket_base_url = get_settings().OUTPUT_BUCKET_BASE_URL
        if using_mock():
            output_bucket_base_url = 'https://mock-bucket'
        if output_bucket_base_url is None:
            raise Exception('Environment variable not set: OUTPUT_BUCKET_BASE_URL')
        output_file_ids = []
        for output_file in job.outputFiles:
            output_file_url = f"{output_bucket_base_url}/dendro-outputs/{job.jobId}/{output_file.name}"
            output_file_id = await _create_output_file(
                file_name=output_file.fileName,
                url=output_file_url,
                project_id=job.projectId,
                user_id=job.userId,
                job_id=job.jobId,
                replace_pending=True,
                is_folder=output_file.isFolder
            )
            output_file_ids.append(output_file_id)
            output_file.fileId = output_file_id
        update['outputFileIds'] = output_file_ids
        update['outputFiles'] = [_model_dump(f, exclude_none=True) for f in job.outputFiles]

    update['status'] = new_status
    if new_error:
        update['error'] = new_error
    if new_status == 'queued':
        update['timestampQueued'] = time.time()
    elif new_status == 'starting':
        update['timestampStarting'] = time.time()
    elif new_status == 'running':
        update['timestampStarted'] = time.time()
    elif new_status == 'completed':
        update['timestampFinished'] = time.time()
    elif new_status == 'failed':
        update['timestampFinished'] = time.time()

    await update_job(job_id=job.jobId, update=update)

    await publish_pubsub_message(
        channel=job.computeResourceId,
        message={
            'type': 'jobStatusChanged',
            'projectId': job.projectId,
            'computeResourceId': job.computeResourceId,
            'jobId': job.jobId,
            'status': new_status
        }
    )



from typing import List, Union
import time
import aiohttp
import uuid
from ..common import AuthException
from ...clients.db import fetch_user_for_dendro_api_key


_user_ids_for_github_access_tokens = {} # github access token -> {user_id: string, timestamp: float}
_user_ids_for_dendro_api_keys = {} # dendro api key -> {user_id: string, timestamp: float}

_mock_github_access_tokens: List[str] = []
def _create_mock_github_access_token():
    token = 'mock:' + str(uuid.uuid4())
    _mock_github_access_tokens.append(token)
    return token

async def _authenticate_gui_request(*,
    github_access_token: Union[str, None] = None,
    dendro_api_key: Union[str, None] = None,
    raise_on_not_authenticated: bool
) -> Union[str, None]:
    try:
        if github_access_token:
            if github_access_token in _user_ids_for_github_access_tokens:
                a = _user_ids_for_github_access_tokens[github_access_token]
                elapsed = time.time() - a['timestamp']
                if elapsed > 60 * 60: # one hour
                    del _user_ids_for_github_access_tokens[github_access_token] # pragma: no cover
                else:
                    return a['user_id']
            if not github_access_token.startswith('mock:'):
                user_id = await _get_user_id_for_access_token(github_access_token)
                if not user_id:
                    raise AuthException('Invalid github access token')
                user_id = 'github|' + user_id # pragma: no cover
            else:
                if github_access_token not in _mock_github_access_tokens:
                    raise AuthException('Invalid mock github access token')
                user_id = 'github|' + github_access_token[len('mock:'):]
            _user_ids_for_github_access_tokens[github_access_token] = {
                'user_id': user_id,
                'timestamp': time.time()
            }
            return user_id
        elif dendro_api_key:
            if dendro_api_key in _user_ids_for_dendro_api_keys:
                a = _user_ids_for_dendro_api_keys[dendro_api_key]
                elapsed = time.time() - a['timestamp']
                if elapsed > 60 * 1: # one minute (the api key might get revoked so we don't want to cache it for too long)
                    del _user_ids_for_dendro_api_keys[dendro_api_key]
                else:
                    return a['user_id']
            user = await fetch_user_for_dendro_api_key(dendro_api_key)
            assert user.dendroApiKey == dendro_api_key
            assert user.userId
            user_id = user.userId
            _user_ids_for_dendro_api_keys[dendro_api_key] = {
                'user_id': user_id,
                'timestamp': time.time()
            }
            return user_id
        else:
            raise AuthException('User is not authenticated')
    except AuthException as e:
        if raise_on_not_authenticated:
            raise e
        else:
            return None

async def _get_user_id_for_access_token(github_access_token: str):
    url = 'https://api.github.com/user'
    headers = {
        'Authorization': f'token {github_access_token}'
    }
    # do async request
    async with aiohttp.ClientSession() as session:
        async with session.get(url, headers=headers) as response:
            if response.status != 200:
                raise AuthException(f'Error getting user ID from github access token: {response.status}')
            data = await response.json() # pragma: no cover
            return data['login'] # pragma: no cover



from ..clients._get_mongo_client import _get_mongo_client
from ..clients._remove_id_field import _remove_id_field
from ...common.dendro_types import DendroFile, DendroJob


async def _remove_detached_files_and_jobs(project_id: str):
    client = _get_mongo_client()
    files_collection = client['dendro']['files']
    jobs_collection = client['dendro']['jobs']

    files = await files_collection.find({
        'projectId': project_id
    }).to_list(length=None) # type: ignore
    jobs = await jobs_collection.find({
        'projectId': project_id,
        'deleted': {'$ne': True}
    }).to_list(length=None) # type: ignore

    for file in files:
        _remove_id_field(file)
    for job in jobs:
        _remove_id_field(job)
    files = [DendroFile(**x) for x in files]
    jobs = [DendroJob(**x) for x in jobs]

    something_changed = True
    while something_changed:
        file_ids = set(x.fileId for x in files)
        job_ids = set(x.jobId for x in jobs)

        job_ids_to_delete = set()
        for job in jobs:
            if any(x not in file_ids for x in job.inputFileIds):
                job_ids_to_delete.add(job.jobId)
            if job.outputFileIds:
                if any(x not in file_ids for x in job.outputFileIds):
                    job_ids_to_delete.add(job.jobId)
        file_ids_to_delete = set()
        for file in files:
            if file.jobId:
                if file.jobId not in job_ids:
                    file_ids_to_delete.add(file.fileId)
        something_changed = False
        if len(job_ids_to_delete) > 0:
            something_changed = True
            for job_id in job_ids_to_delete:
                await jobs_collection.update_one({
                    'jobId': job_id
                }, {
                    '$set': {
                        'deleted': True
                    }
                })
            jobs = [x for x in jobs if x.jobId not in job_ids_to_delete]
        if len(file_ids_to_delete) > 0:
            something_changed = True
            await files_collection.delete_many({
                'fileId': {'$in': list(file_ids_to_delete)}
            })
            files = [x for x in files if x.fileId not in file_ids_to_delete]



from .Project import Project, load_project # noqa: F401
from .submit_job import submit_job, SubmitJobInputFile, SubmitJobOutputFile, SubmitJobParameter # noqa: F401
from ..common.dendro_types import DendroJobRequiredResources # noqa: F401



# flake8: noqa

from .InputFile import InputFile
from .InputFolder import InputFolder
from .OutputFile import OutputFile
from .OutputFolder import OutputFolder
from .App import App

from .ProcessorBase import ProcessorBase

from .. import BaseModel, Field



from setuptools import setup, find_packages

# read version from dendro/version.txt
with open('dendro/version.txt') as f:
    __version__ = f.read().strip()

setup(
    name='dendro',
    version=__version__,
    author="Jeremy Magland, Luiz Tauffer",
    author_email="jmagland@flatironinstitute.org",
    url="https://github.com/flatironinstitute/dendro",
    description="Web framework for neurophysiology data analysis",
    packages=find_packages(),
    include_package_data=True,
    package_data={'dendro': ['version.txt']},
    install_requires=[
        'click',
        'simplejson',
        'numpy',
        'PyYAML',
        'remfile',
        'pydantic', # intentionally do not specify version 1 or 2 since we support both
        'cryptography',
        'h5py>=3.10.0',
        'psutil'
    ],
    extras_require={
        'compute_resource': [
            'pubnub>=7.2.0',
            'boto3'
        ],
        'api': [
            'fastapi',
            'motor',
            'simplejson',
            'pydantic',
            'aiohttp',
            'boto3'
        ]
    },
    entry_points={
        "console_scripts": [
            "dendro=dendro.cli:main",
        ],
    }
)



import os
import time
import yaml
from typing import Optional, Tuple
from ..common._crypto_keys import sign_message, generate_keypair


env_var_keys = [
    'COMPUTE_RESOURCE_ID',
    'COMPUTE_RESOURCE_PRIVATE_KEY',
    'CONTAINER_METHOD',
    'DEFAULT_JOB_RUN_METHOD',
    'AVAILABLE_JOB_RUN_METHODS',
    'AWS_ACCESS_KEY_ID',
    'AWS_DEFAULT_REGION',
    'AWS_SECRET_ACCESS_KEY'
]

def register_compute_resource(*, dir: str, compute_resource_id: Optional[str] = None, compute_resource_private_key: Optional[str] = None) -> Tuple[str, str]:
    """Initialize a Dendro compute resource node.

    Args:
        dir: The directory associated with the compute resource node.
    """

    # Let's make sure pubnub is installed, because it's required for the daemon
    try:
        import pubnub # noqa
    except ImportError:
        raise ImportError('The pubnub package is not installed. You should use "pip install dendro[compute_resource]".')

    env_fname = os.path.join(dir, '.dendro-compute-resource-node.yaml')

    the_env = {}
    for k in env_var_keys:
        the_env[k] = ''
    if not os.path.exists(env_fname):
        if not compute_resource_id:
            if compute_resource_private_key:
                raise ValueError('Cannot specify compute_resource_private_key without specifying compute_resource_id.')
            public_key_hex, private_key_hex = generate_keypair()
            compute_resource_id = public_key_hex
            compute_resource_private_key = private_key_hex
        else:
            if not compute_resource_private_key:
                raise ValueError('Cannot specify compute_resource_id without specifying compute_resource_private_key.')
        the_env['COMPUTE_RESOURCE_ID'] = compute_resource_id
        the_env['COMPUTE_RESOURCE_PRIVATE_KEY'] = compute_resource_private_key
        the_env['CONTAINER_METHOD'] = os.getenv('CONTAINER_METHOD', '')

        with open(env_fname, 'w', encoding='utf8') as f:
            yaml.dump(the_env, f)
    elif compute_resource_id is not None or compute_resource_private_key is not None:
        raise ValueError('Cannot specify compute_resource_id or compute_resource_private_key if compute resource node is already initialized.')

    with open(env_fname, 'r', encoding='utf8') as f:
        the_env_from_config_file = yaml.safe_load(f)

    for k in env_var_keys:
        if k in the_env_from_config_file:
            the_env[k] = the_env_from_config_file[k]
        else:
            the_env_from_config_file[k] = ''

    with open(env_fname, 'w', encoding='utf8') as f:
        yaml.dump(the_env_from_config_file, f)

    COMPUTE_RESOURCE_ID = the_env['COMPUTE_RESOURCE_ID']
    COMPUTE_RESOURCE_PRIVATE_KEY = the_env['COMPUTE_RESOURCE_PRIVATE_KEY']

    if not COMPUTE_RESOURCE_ID:
        raise ValueError('COMPUTE_RESOURCE_ID is not set')
    if not COMPUTE_RESOURCE_PRIVATE_KEY:
        raise ValueError('COMPUTE_RESOURCE_PRIVATE_KEY is not set')

    timestamp = int(time.time())
    msg = {
        'timestamp': timestamp
    }
    signature = sign_message(msg, COMPUTE_RESOURCE_ID, COMPUTE_RESOURCE_PRIVATE_KEY)
    resource_code = f'{timestamp}-{signature}'

    url = f'https://dendro.vercel.app/register-compute-resource/{COMPUTE_RESOURCE_ID}/{resource_code}'
    print('')
    print('Please visit the following URL in your browser to register your compute resource:')
    print('')
    print(url)
    print('')

    assert COMPUTE_RESOURCE_ID is not None
    assert COMPUTE_RESOURCE_PRIVATE_KEY is not None
    return COMPUTE_RESOURCE_ID, COMPUTE_RESOURCE_PRIVATE_KEY
